# labels: test_group::turnkey name::generalconv author::graph_convolutions task::Graph_Machine_Learning license::mit
"""
A general GNN layer adapted from the `"Design Space for Graph Neural Networks"
<https://arxiv.org/abs/2011.08843>`_ paper.
"""

import torch

from torch_geometric.datasets import Planetoid
from torch_geometric.nn import GeneralConv

dataset = Planetoid(root=".", name="Cora")
data = dataset[0]
edge_index_rows = 2


model = GeneralConv(dataset.num_features, dataset.num_classes)
x = torch.ones(data.num_nodes, data.num_features, dtype=torch.float)
edge_index = torch.ones(edge_index_rows, data.num_nodes, dtype=torch.long)
inputs = {
    "x": x,
    "edge_index": edge_index,
}

# Call model
model(**inputs)



import os
import csv
from datetime import datetime
from pathlib import Path
from typing import Dict, List
import yaml
import pandas as pd
import turnkeyml.common.printing as printing
import turnkeyml.common.filesystem as fs
import turnkeyml.common.build as bd


def get_report_name(prefix: str = "") -> str:
    """
    Returns the name of the .csv report
    """
    day = datetime.now().day
    month = datetime.now().month
    year = datetime.now().year
    date_key = f"{year}-{str(month).zfill(2)}-{str(day).zfill(2)}"
    return f"{prefix}{date_key}.csv"


def _good_get(
    dict: Dict, key: str, return_keys: bool = False, return_values: bool = False
):
    if key in dict:
        if return_keys:
            return list(dict[key].keys())
        elif return_values:
            return list(dict[key].values())
        else:
            return dict[key]
    else:
        return "-"


def summary_spreadsheets(args) -> None:
    # Input arguments from CLI
    cache_dirs = [os.path.expanduser(dir) for dir in args.cache_dirs]
    cache_dirs = fs.expand_inputs(cache_dirs)
    report_dir = os.path.expanduser(args.report_dir)

    # Name report file
    report_path = os.path.join(report_dir, get_report_name())

    # Create report dict
    Path(report_dir).mkdir(parents=True, exist_ok=True)

    report: List[Dict] = []
    all_evaluation_stats = []

    # Add results from all user-provided cache folders
    for cache_dir in cache_dirs:
        # Check if this is a valid cache directory
        fs.check_cache_dir(cache_dir)

        # List all yaml files available
        all_model_stats_yamls = fs.get_all(
            path=cache_dir, file_type="turnkey_stats.yaml"
        )
        all_model_stats_yamls = sorted(all_model_stats_yamls)

        # Bring all of the stats for all of the models into memory
        for model_stats_yaml in all_model_stats_yamls:
            with open(model_stats_yaml, "r", encoding="utf8") as stream:
                try:
                    # load the yaml into a dict
                    model_stats = yaml.load(stream, Loader=yaml.FullLoader)

                    # create a separate dict for each evaluation
                    for evaluation in model_stats[fs.Keys.EVALUATIONS].values():
                        evaluation_stats = {}

                        # Copy all of the stats for the model that are common across evaluation
                        for key, value in model_stats.items():
                            if key != fs.Keys.EVALUATIONS:
                                evaluation_stats[key] = value

                        # Copy the evaluation-specific stats
                        for key, value in evaluation.items():
                            # If a build or benchmark is still marked as "incomplete" at
                            # reporting time, it must have been killed by a time out,
                            # out-of-memory (OOM), or some other uncaught exception
                            if (
                                (
                                    key == fs.Keys.BUILD_STATUS
                                    or fs.Keys.BENCHMARK_STATUS
                                )
                                or fs.Keys.STAGE_STATUS in key
                            ) and value == bd.FunctionStatus.INCOMPLETE.value:
                                value = bd.FunctionStatus.KILLED.value

                            # Add stats ensuring that those are all in lower case
                            evaluation_stats[key.lower()] = value

                        all_evaluation_stats.append(evaluation_stats)
                except yaml.scanner.ScannerError:
                    continue

    # Scan the build stats to determine the set of columns for the CSV file.
    # The CSV will have one column for every key in any build stats dict.
    column_headers = []
    for evaluation_stats in all_evaluation_stats:
        # Add any key that isn't already in column_headers
        for header in evaluation_stats.keys():
            if header not in column_headers:
                column_headers.append(header)

    # Sort all columns alphabetically
    column_headers = sorted(column_headers)

    # Add each build to the report
    for evaluation_stats in all_evaluation_stats:
        # Start with a dictionary where all of the values are "-". If a build
        # has a value for each key we will fill it in, and otherwise the "-"
        # will indicate that no value was available
        result = {k: "-" for k in column_headers}

        for key in column_headers:
            result[key] = _good_get(evaluation_stats, key)

        report.append(result)

    # Populate results spreadsheet
    with open(report_path, "w", newline="", encoding="utf8") as spreadsheet:
        writer = csv.writer(spreadsheet)
        writer.writerow(column_headers)
        for build in report:
            writer.writerow([build[col] for col in column_headers])

    # Print message with the output file path
    printing.log("Summary spreadsheet saved at ")
    printing.logn(str(report_path), printing.Colors.OKGREEN)

    # Save the unique errors and counts to a file
    errors = []
    for evaluation_stats in all_evaluation_stats:
        if (
            "compilation_error" in evaluation_stats.keys()
            and "compilation_error_id" in evaluation_stats.keys()
        ):
            error = evaluation_stats["compilation_error"]
            id = evaluation_stats["compilation_error_id"]
            if id != "":
                unique_error = True
                for reported_error in errors:
                    if reported_error["id"] == id:
                        unique_error = False
                        reported_error["count"] = reported_error["count"] + 1
                        reported_error["models_impacted"] = reported_error[
                            "models_impacted"
                        ] + [evaluation_stats["model_name"]]

                if unique_error:
                    reported_error = {
                        "id": id,
                        "count": 1,
                        "models_impacted": [evaluation_stats["model_name"]],
                        "example": error,
                    }
                    errors.append(reported_error)

    if len(errors) > 0:
        errors_path = os.path.join(report_dir, get_report_name("errors-"))
        with open(errors_path, "w", newline="", encoding="utf8") as spreadsheet:
            writer = csv.writer(spreadsheet)
            error_headers = errors[0].keys()
            writer.writerow(error_headers)
            for unique_error in errors:
                writer.writerow([unique_error[col] for col in error_headers])

        printing.log("Compilation errors spreadsheet saved at ")
        printing.logn(str(errors_path), printing.Colors.OKGREEN)
    else:
        printing.logn(
            "No compilation errors in any cached build, skipping errors spreadsheet."
        )


def get_dict(report_csv: str, columns: List[str]) -> Dict[str, Dict[str, str]]:
    """
    Returns a dictionary where the keys are model names and the values are dictionaries.
    Each dictionary represents a model with column names as keys and their corresponding values.
    args:
     - report_csv: path to a report.csv file generated by turnkey CLI
     - columns: list of column names in the report.csv file whose values will be used to
        populate the dictionary
    """

    # Load the report as a dataframe
    dataframe = pd.read_csv(report_csv)

    # Create a nested dictionary with model_name as keys and another
    # dictionary of {column: value} pairs as values
    result = {
        row[0]: row[1].to_dict()
        for row in dataframe.set_index("model_name")[columns].iterrows()
    }

    return result



import os
import pathlib
import json
import shutil
import subprocess
from typing import Dict, Any, Optional, List, Union
from abc import ABC, abstractmethod
import torch
import numpy as np
from turnkeyml.common.performance import MeasuredPerformance, Device
import turnkeyml.common.build as build
import turnkeyml.common.exceptions as exp
from turnkeyml.common.filesystem import Stats, rebase_cache_dir


def _check_docker_install():
    # Check if docker and python are installed on the local machine
    docker_location = shutil.which("docker")
    if not docker_location:
        raise ValueError("Docker installation not found. Please install Docker>=20.10")


def _check_docker_running():
    try:
        # On both Windows and Linux, 'docker info' will return an error if the
        #  Docker daemon is not running
        subprocess.run(
            ["docker", "info"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError:
        raise ValueError(
            "Docker engine is not running. Please start the Docker engine."
        )


class BaseRT(ABC):
    def __init__(
        self,
        cache_dir: str,
        build_name: str,
        stats: Stats,
        device_type: Union[str, Device],
        runtime: str,
        runtimes_supported: List[str],
        runtime_version: str,
        base_path: str,
        iterations: int = 100,
        model: Optional[torch.nn.Module] = None,
        inputs: Optional[Dict[str, Any]] = None,
        requires_docker: bool = False,
        tensor_type=np.array,
        execute_function: Optional[callable] = None,
    ):
        self.tensor_type = tensor_type
        self.cache_dir = cache_dir
        self.build_name = build_name
        self.stats = stats
        if isinstance(device_type, Device):
            self.device_type = device_type
        else:
            self.device_type = Device(device_type)
        self.runtime = runtime
        self.iterations = iterations
        self.requires_docker = requires_docker
        self.conda_script = os.path.join(base_path, "within_conda.py")
        self.conda_requirements = os.path.join(base_path, "requirements.txt")
        self.docker_output_dir = "/app"
        self.runtime_version = runtime_version
        self.model = model
        self.inputs = inputs
        self.onnx_filename = "model.onnx"
        self.onnx_dirname = "onnxmodel"
        self.outputs_filename = "outputs.json"
        self.runtimes_supported = runtimes_supported
        self.execute_function = execute_function

        # Validate runtime is supported
        if runtime not in runtimes_supported:
            raise ValueError(
                f"'runtime' argument {runtime} passed to TensorRT, which only "
                f"supports runtimes: {runtimes_supported}"
            )

        self._setup()

    def posix_path_format(self, path) -> str:
        """
        Reformat a path into posix format. Necessary for when a path is
        created with os.path.join() on Windows that will be used in a posix
        environment such as a Docker container.
        """
        return pathlib.Path(path).as_posix()

    @property
    def local_output_dir(self):
        return os.path.join(
            build.output_dir(self.cache_dir, self.build_name),
            f"{str(self.device_type).replace('::','_')}_benchmark",
        )

    @property
    def local_onnx_dir(self):
        return os.path.join(self.local_output_dir, self.onnx_dirname)

    @property
    def docker_onnx_dir(self):
        return self.posix_path_format(
            os.path.join(self.docker_output_dir, self.onnx_dirname)
        )

    @property
    def local_onnx_file(self):
        return os.path.join(self.local_onnx_dir, self.onnx_filename)

    @property
    def docker_onnx_file(self):
        return self.posix_path_format(
            os.path.join(self.docker_onnx_dir, self.onnx_filename)
        )

    @property
    def local_outputs_file(self):
        return os.path.join(self.local_output_dir, self.outputs_filename)

    @property
    def docker_outputs_file(self):
        return self.posix_path_format(
            os.path.join(self.docker_output_dir, self.outputs_filename)
        )

    def _transfer_files(self, files_to_transfer: List[str]):
        """
        output_dir: destination for files
        files_to_transfer: absolute paths to files
        """

        os.makedirs(self.local_output_dir, exist_ok=True)
        for file in files_to_transfer:
            shutil.copy(
                file, os.path.join(self.local_output_dir, os.path.basename(file))
            )

    @abstractmethod
    def _setup(self) -> None:
        """
        Any code that should be called prior to benchmarking as a one-time setup.
        Called automatically at the end of  `BaseRT.__init__()`.
        """

    def _execute(self, output_dir, onnx_file, outputs_file) -> None:
        """
        Execute the benchmark in hardware using the selected runtime. Must produce output
        artifacts that can be parsed by the other methods of this class.
        Overloading this method is optional; the default implementation of benchmark()
        calls this method, however you are free to overload benchmark() to behave differently.
        """
        raise NotImplementedError

    def benchmark(self) -> MeasuredPerformance:
        """
        Transfer input artifacts, execute model on hardware, analyze output artifacts,
        and return the performance.
        """

        # Remove previous benchmarking artifacts
        if os.path.exists(self.local_outputs_file):
            os.remove(self.local_outputs_file)

        # Transfer input artifacts
        state = build.load_state(self.cache_dir, self.build_name)

        # Just in case the model file was generated on a different machine:
        # strip the state's cache dir, then prepend the current cache dir
        model_file = rebase_cache_dir(
            state.results[0], state.config.build_name, self.cache_dir
        )

        if not os.path.exists(model_file):
            msg = "Model file not found"
            raise exp.ModelRuntimeError(msg)

        os.makedirs(self.local_output_dir, exist_ok=True)
        os.makedirs(self.local_onnx_dir, exist_ok=True)
        shutil.copy(model_file, self.local_onnx_file)

        # Execute benchmarking in hardware
        if self.requires_docker:
            _check_docker_install()
            onnx_file = self.docker_onnx_file
            _check_docker_running()
        else:
            onnx_file = self.local_onnx_file

        self._execute(
            output_dir=self.local_output_dir,
            onnx_file=onnx_file,
            outputs_file=self.local_outputs_file,
        )

        if not os.path.isfile(self.local_outputs_file):
            raise exp.BenchmarkException(
                "No benchmarking outputs file found after benchmarking run. "
                "Sorry we don't have more information."
            )

        # Call property methods to analyze the output artifacts for performance stats
        # and return them
        return MeasuredPerformance(
            mean_latency=self.mean_latency,
            throughput=self.throughput,
            device=self.device_name,
            device_type=self.device_type,
            runtime=self.runtime,
            runtime_version=self.runtime_version,
            build_name=self.build_name,
        )

    def _get_stat(self, stat):
        if os.path.exists(self.local_outputs_file):
            with open(self.local_outputs_file, encoding="utf-8") as f:
                performance = json.load(f)
            return performance[stat]
        else:
            raise exp.BenchmarkException(
                "No benchmarking outputs file found after benchmarking run."
                "Sorry we don't have more information."
            )

    @property
    @abstractmethod
    def mean_latency(self) -> float:
        """
        Returns the mean latency, in ms, for the benchmarking run.
        """

    @property
    @abstractmethod
    def throughput(self) -> float:
        """
        Returns the throughput, in IPS, for the benchmarking run.
        """

    @property
    @abstractmethod
    def device_name(self) -> str:
        """
        Returns the full device name for the device used in benchmarking.
        For example, a benchmark on a `x86` device might have a device name like
        `AMD Ryzen 7 PRO 6850U with Radeon Graphics`.
        """



import abc
import sys
import time
import os
import copy
from typing import List, Tuple
from multiprocessing import Process
import psutil
import turnkeyml.common.printing as printing
import turnkeyml.common.exceptions as exp
import turnkeyml.common.build as build
import turnkeyml.common.filesystem as fs


def _spinner(message):
    parent_pid = os.getppid()
    while psutil.pid_exists(parent_pid):
        for cursor in ["   ", ".  ", ".. ", "..."]:
            time.sleep(0.5)
            status = f"      {message}{cursor}\r"
            sys.stdout.write(status)
            sys.stdout.flush()


def _name_is_file_safe(name: str):
    """
    Make sure the name can be used in a filename
    """

    allowed_in_unique_name = set(
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"
    )

    if len(name) == 0:
        msg = """
        Stage __init__() was passed a unique_name with no length. A
        uniquely identifying unique_name is required.
        """
        raise ValueError(msg)

    for char in name:
        if char not in allowed_in_unique_name:
            msg = f"""
            Stage __init__() was passed a unique_name:
            {name}
            with illegal characters. The unique_name must be safe to
            use in a filename, meaning it can only use characters: {allowed_in_unique_name}
            """
            raise ValueError(msg)


class Stage(abc.ABC):
    def status_line(self, successful, verbosity):
        """
        Print a line of status information for this Stage into the monitor.
        """
        if verbosity:
            # Only use special characters when the terminal encoding supports it
            if sys.stdout.encoding == "utf-8":
                success_tick = "✓"
                fail_tick = "×"
            else:
                success_tick = "+"
                fail_tick = "x"

            if successful is None:
                # Initialize the message
                printing.logn(f"      {self.monitor_message}   ")
            elif successful:
                # Print success message
                printing.log(f"    {success_tick} ", c=printing.Colors.OKGREEN)
                printing.logn(self.monitor_message + "   ")
            else:
                # successful == False, print failure message
                printing.log(f"    {fail_tick} ", c=printing.Colors.FAIL)
                printing.logn(self.monitor_message + "   ")

    def __init__(
        self,
        unique_name,
        monitor_message,
    ):
        _name_is_file_safe(unique_name)

        self.unique_name = unique_name
        self.status_key = f"{fs.Keys.STAGE_STATUS}:{unique_name}"
        self.duration_key = f"{fs.Keys.STAGE_DURATION}:{unique_name}"
        self.monitor_message = monitor_message
        self.progress = None
        self.logfile_path = None
        self.stages = None

    @abc.abstractmethod
    def fire(self, state: build.State) -> build.State:
        """
        Developer-defined function to fire the stage.
        In less punny terms, this is the function that
        build_model() will run to implement a model-to-model
        transformation on the flow to producing a Model.
        """

    def fire_helper(self, state: build.State) -> Tuple[build.State, int]:
        """
        Wraps the user-defined .fire method with helper functionality.
        Specifically:
            - Provides a path to a log file
            - Redirects the stdout of the stage to that log file
            - Monitors the progress of the stage on the command line,
                including in the event of an exception
        """

        # Set the build status to INCOMPLETE to indicate that a Stage
        # started running. This allows us to test whether the Stage exited
        # unexpectedly, before it was able to set ERROR
        state.build_status = build.FunctionStatus.INCOMPLETE

        self.logfile_path = os.path.join(
            build.output_dir(state.cache_dir, state.config.build_name),
            f"log_{self.unique_name}.txt",
        )

        if state.monitor:
            self.progress = Process(target=_spinner, args=[self.monitor_message])
            self.progress.start()

        try:
            # Execute the build stage
            with build.Logger(self.monitor_message, self.logfile_path):
                state = self.fire(state)

        except exp.StageError:
            self.status_line(
                successful=False,
                verbosity=state.monitor,
            )
            state.build_status = build.FunctionStatus.ERROR
            raise

        else:
            self.status_line(successful=True, verbosity=state.monitor)

            # Stages should not set build.FunctionStatus.SUCCESSFUL for the whole build,
            # as that is reserved for Sequence.launch()
            if state.build_status == build.FunctionStatus.SUCCESSFUL:
                raise exp.StageError(
                    "TurnkeyML Stages are not allowed to set "
                    "`state.build_status == build.FunctionStatus.SUCCESSFUL`, "
                    "however that has happened. If you are a plugin developer, "
                    "do not do this. If you are a user, please file an issue at "
                    "https://github.com/onnx/turnkeyml/issues."
                )

        finally:
            if state.monitor:
                self.progress.terminate()

        return state

    def get_names(self) -> List[str]:
        """
        Sequence uses self.names() to recursively get the names of all
        Stages in the Sequence. An individual Stage just needs to return
        its own name.
        """
        if self.stages is None:
            return [self.unique_name]
        else:
            result = []
            for stage in self.stages:
                result = result + stage.get_names()

            return result

    def get_depth(self) -> int:
        """
        Sequence needs to know the depth of each Stage within the Sequence in order
        to properly update the terminal UI. An individual Stage just needs to return
        the value 1.
        """
        if self.stages is None:
            return 1
        else:
            count = 0
            for stage in self.stages:
                count = count + stage.get_depth()
            return count


def _rewind_stdout(lines: int):
    """
    Helper function for the command line monitor. Moves the cursor up a
    certain number of lines in the terminal, corresponding to the
    status line for a Stage, so that we can update the status of
    that Stage.
    """
    rewind_stdout_one_line = "\033[1A"
    rewind_multiple_lines = rewind_stdout_one_line * lines
    print(rewind_multiple_lines, end="")


def unroll_stages(stages):
    """
    Recursively goes through all sequences and returns list of stages
    """

    unrolled_stages = []
    for stage in stages:
        if isinstance(stage, Sequence):
            unrolled_stages += unroll_stages(stage.stages)
        else:
            unrolled_stages += [stage]
    return unrolled_stages


class Sequence(Stage):
    def __init__(
        self,
        unique_name,
        monitor_message,
        stages: List[Stage],
        enable_model_validation=False,
    ):
        super().__init__(unique_name, monitor_message)

        # The `stages` argument can be a nested Sequence (ie, Sequence of Sequence of Stage).
        # Unroll the stages to make the Sequence easier to deal with
        self.stages = unroll_stages(stages)

        # Follow default model validation steps in ignition.model_intake()
        self.enable_model_validation = enable_model_validation

        # Make sure all the stage names are unique
        stage_names = self.get_names()

        if len(stage_names) != len(set(stage_names)):
            msg = f"""
            All Stages in a Sequence must have unique unique_names, however Sequence
            received duplicates in the list of names: {stage_names}
            """
            raise ValueError(msg)

    def show_monitor(self, config: build.Config, verbosity: bool):
        """
        Displays the monitor on the terminal. The purpose of the monitor
        is to show the status of each stage (success, failure, not started yet,
        or in-progress).
        """

        if verbosity:
            print("\n\n")

            printing.logn(
                f'Building "{config.build_name}"',
                c=printing.Colors.BOLD,
            )

            for stage in self.stages:
                stage.status_line(successful=None, verbosity=True)

            _rewind_stdout(self.get_depth())

    def launch(self, state: build.State) -> build.State:
        """
        Executes a launch sequence.
        In less punny terms, this method is called by the top-level
        build_model() function to iterate over all of the Stages required for a build.
        Builds are defined by self.stages in a top-level Sequence, and self.stages
        can include both Stages and Sequences (ie, sequences can be nested).
        """

        if state.build_status == build.FunctionStatus.SUCCESSFUL:
            msg = """
            build_model() is running a build on a model that already built successfully, which
            should not happen because the build should have loaded from cache or rebuilt from scratch.
            If you are using custom Stages and Sequences then you have some debugging to do. Otherwise,
            please file an issue at https://github.com/onnx/turnkeyml/issues
            """
            raise exp.Error(msg)

        # Collect telemetry for the build
        stats = fs.Stats(state.cache_dir, state.config.build_name, state.evaluation_id)
        stats.save_model_eval_stat(
            fs.Keys.SELECTED_SEQUENCE_OF_STAGES,
            self.get_names(),
        )

        # At the beginning of a sequence no stage has started
        for stage in self.stages:
            stats.save_model_eval_stat(
                stage.status_key, build.FunctionStatus.NOT_STARTED.value
            )
            stats.save_model_eval_stat(stage.duration_key, "-")

        # Run the build
        start_time = time.time()
        try:
            for stage in self.stages:
                # Set status as incomplete, since stage just started
                stats.save_model_eval_stat(
                    stage.status_key, build.FunctionStatus.INCOMPLETE.value
                )

                # Collect telemetry about the stage
                state.current_build_stage = stage.unique_name

                # Run the stage
                state = stage.fire_helper(state)

                # Set status as successful
                stats.save_model_eval_stat(
                    stage.status_key, build.FunctionStatus.SUCCESSFUL.value
                )

        except exp.StageError as e:
            # Advance the cursor below the monitor so
            # we can print an error message
            stage_depth_in_sequence = self.get_depth() - self.get_names().index(
                stage.unique_name  # pylint: disable=undefined-loop-variable
            )
            stdout_lines_to_advance = stage_depth_in_sequence - 2
            cursor_down = "\n" * stdout_lines_to_advance

            print(cursor_down)

            printing.log_error(e)

            stats.save_model_eval_stat(
                stage.status_key, build.FunctionStatus.ERROR.value
            )

            raise

        else:
            state.current_build_stage = None
            state.build_status = build.FunctionStatus.SUCCESSFUL

            # We use a deepcopy here because the Stage framework supports
            # intermediate_results of any type, including model objects in memory.
            # The deepcopy ensures that we are providing a result that users
            # are free to take any action with.
            state.results = copy.deepcopy(state.intermediate_results)

            return state

        finally:
            # Collect telemetry about the stage
            execution_time = time.time() - start_time
            stats.save_model_eval_stat(stage.duration_key, execution_time)

    def status_line(self, successful, verbosity):
        """
        This override of status_line simply propagates status_line()
        to every Stage in the Sequence
        FIXME: A cleaner implementation of Stage/Sequence might not need this
        """
        for stage in self.stages:
            stage.status_line(successful=None, verbosity=verbosity)

    def fire(self, state: build.State) -> build.State:
        """
        This override of fire simply propagates fire()
        to every Stage in the Sequence
        FIXME: A cleaner implementation of Stage/Sequence might not need this
        """
        for stage in self.stages:
            state = stage.fire_helper(state)

        return state

    def fire_helper(self, state: build.State) -> build.State:
        """
        Sequence doesn't need any help calling self.fire(), so it's fire_helper
        is just to call self.fire()
        FIXME: A cleaner implementation of Stage/Sequence might not need this
        """
        return self.fire(state)



# labels: test_group::turnkey name::midas_v3_large author::torch_hub task::Computer_Vision license::bsd-3-clause
"""https://pytorch.org/hub/intelisl_midas_v2/"""
from turnkeyml.parser import parse
import torch

torch.manual_seed(0)

# Parsing command-line arguments
batch_size, height, num_channels, width = parse(
    ["batch_size", "height", "num_channels", "width"]
)


# Model and input configurations
model = torch.hub.load("intel-isl/MiDaS", "DPT_Large")

inputs = {"x": torch.ones(batch_size, num_channels, height, width, dtype=torch.float)}


# Call model
model(**inputs)



# labels: name::albert_large_v1 author::transformers task::Generative_AI license::apache-2.0
from turnkeyml.parser import parse
from transformers import AlbertModel, AutoConfig
import torch

torch.manual_seed(0)

# Parsing command-line arguments
pretrained, batch_size, max_seq_length = parse(
    ["pretrained", "batch_size", "max_seq_length"]
)

# Model and input configurations
if pretrained:
    model = AlbertModel.from_pretrained("albert-large-v1")
else:
    config = AutoConfig.from_pretrained("albert-large-v1")
    model = AlbertModel(config)

# Make sure the user's sequence length fits within the model's maximum
assert max_seq_length <= model.config.max_position_embeddings


inputs = {
    "input_ids": torch.ones(batch_size, max_seq_length, dtype=torch.long),
    "attention_mask": torch.ones(batch_size, max_seq_length, dtype=torch.float),
}


# Call model
model(**inputs)



"""
Utilities for spawning new turnkey calls, in both local processes and on Slurm
"""

import os
import subprocess
import pathlib
import time
import shlex
import platform
import getpass
from typing import List, Optional, Dict, Union
from enum import Enum
import turnkeyml.common.filesystem as filesystem
import turnkeyml.common.printing as printing
from turnkeyml.cli.parser_helpers import encode_args

if os.environ.get("TURNKEY_TIMEOUT_SECONDS"):
    timeout_env_var = os.environ.get("TURNKEY_TIMEOUT_SECONDS")
    SECONDS_IN_A_DAY = 60 * 60 * 24
    if timeout_env_var > SECONDS_IN_A_DAY:
        raise ValueError(
            f"Value of TURNKEY_TIMEOUT_SECONDS must be less than 1 day, got {timeout_env_var}"
        )
    DEFAULT_TIMEOUT_SECONDS = int(timeout_env_var)
else:
    DEFAULT_TIMEOUT_SECONDS = 3600


class Target(Enum):
    SLURM = "slurm"
    LOCAL_PROCESS = "local_process"


def slurm_jobs_in_queue(job_name=None) -> List[str]:
    """Return the set of slurm jobs that are currently pending/running"""
    user = getpass.getuser()
    if job_name is None:
        output = subprocess.check_output(["squeue", "-u", user])
    else:
        output = subprocess.check_output(["squeue", "-u", user, "--name", job_name])
    output = output.split(b"\n")
    output = [s.decode("utf").split() for s in output]

    # Remove headers
    output.pop(0)

    # Remove empty line at the end
    output.pop(-1)

    # Get just the job names
    if len(output) > 0:
        name_index_in_squeue = 2
        output = [s[name_index_in_squeue] for s in output]

    return output


def arg_format(name: str):
    name_underscores = name.replace("_", "-")
    return f"--{name_underscores}"


def list_arg(key: str, values: List):
    if values is not None:
        result = " ".join(values)
        return f"{key} {result}"
    else:
        return ""


def value_arg(key: str, value: Union[str, int]):
    if value is not None:
        return f'{key}="{value}"'
    else:
        return ""


def bool_arg(key: str, value: bool):
    if value:
        return f"{key}"
    else:
        return ""


def dict_arg(key: str, value: Dict):
    if value:
        return f"{key} {' '.join(encode_args(value))}"
    else:
        return ""


def run_turnkey(
    op: str,
    file_name: str,
    target: Target,
    timeout: Optional[int] = DEFAULT_TIMEOUT_SECONDS,
    working_dir: str = os.getcwd(),
    ml_cache_dir: Optional[str] = os.environ.get("SLURM_ML_CACHE"),
    max_jobs: int = 50,
    **kwargs,
):
    """
    Run turnkey on a single input file in a separate process (e.g., Slurm, subprocess).
    Any arguments that should be passed to the new turnkey process must be provided via kwargs.

    kwargs must also match the following format:
      The key must be the snake_case version of the CLI argument (e.g, build_only for --build-only)
    """

    type_to_formatter = {
        str: value_arg,
        int: value_arg,
        bool: bool_arg,
        list: list_arg,
        dict: dict_arg,
    }

    invocation_args = f"{op} {file_name}"

    for key, value in kwargs.items():
        if value is not None:
            arg_str = type_to_formatter[type(value)](arg_format(key), value)
            invocation_args = invocation_args + " " + arg_str

    if target == Target.SLURM:
        # Change args into the format expected by Slurm
        slurm_args = " ".join(shlex.split(invocation_args))

        # Remove the .py extension from the build name
        job_name = filesystem.clean_file_name(file_name)

        # Put the timeout into format days-hh:mm:ss
        hh_mm_ss = time.strftime("%H:%M:%S", time.gmtime(timeout))
        slurm_format_timeout = f"00-{hh_mm_ss}"

        while len(slurm_jobs_in_queue()) >= max_jobs:
            print(
                f"Waiting: Your number of jobs running ({len(slurm_jobs_in_queue())}) "
                "matches or exceeds the maximum "
                f"concurrent jobs allowed ({max_jobs}). "
                f"The jobs in queue are: {slurm_jobs_in_queue()}"
            )
            time.sleep(5)

        shell_script = os.path.join(
            pathlib.Path(__file__).parent.resolve(), "run_slurm.sh"
        )

        slurm_command = ["sbatch", "-c", "1"]
        if os.environ.get("TURNKEY_SLURM_USE_DEFAULT_MEMORY") != "True":
            slurm_command.append("--mem=128000")
        slurm_command.extend(
            [
                f"--time={slurm_format_timeout}",
                f"--job-name={job_name}",
                shell_script,
                "turnkey",
                slurm_args,
                working_dir,
            ]
        )
        if ml_cache_dir is not None:
            slurm_command.append(ml_cache_dir)

        print(f"Submitting job {job_name} to Slurm")
        subprocess.check_call(slurm_command)
    elif target == Target.LOCAL_PROCESS:
        command = "turnkey " + invocation_args
        printing.log_info(f"Starting process with command: {command}")

        # Linux and Windows want to handle some details differently
        if platform.system() != "Windows":
            command = shlex.split(command)

        # Launch a subprocess for turnkey to evaluate the script
        try:
            subprocess.check_call(command, stderr=subprocess.STDOUT, timeout=timeout)
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as e:
            printing.log_error(
                "Process was terminated with the error shown below. "
                f"turnkey will move on to the next input.\n{e}"
            )

            if "Signals.SIGKILL: 9" in str(e):
                printing.log_info(
                    "It is possible your computer ran out of memory while attempting to evaluate "
                    f"{file_name}. You can check this by repeating the experiment "
                    "while using `top` to monitor memory utilization."
                )
    else:
        raise ValueError(f"Unsupported value for target: {target}.")



import subprocess
import os
import threading
import json
import numpy as np
from turnkeyml.run.tensorrt.execute import TRT_VERSION
from turnkeyml.run.basert import BaseRT
import turnkeyml.common.exceptions as exp
from turnkeyml.common.filesystem import Stats
from turnkeyml.run.tensorrt.execute import (
    measure_power,
    run,
    average_power_and_utilization,
)


def _get_nvidia_driver_version():
    try:
        output = subprocess.check_output(["nvidia-smi"], text=True)

        # Search for the driver version in the output
        for line in output.split("\n"):
            if "Driver Version" in line:
                # Extract and return the driver version
                return line.split(":")[1].strip().split()[0]

    except Exception as e:  # pylint: disable=broad-except
        return str(e)

    return "Driver not found"


class TensorRT(BaseRT):
    def __init__(
        self,
        cache_dir: str,
        build_name: str,
        stats: Stats,
        iterations: int,
        device_type: str,
        runtime: str = "trt",
        tensor_type=np.array,
        model=None,
        inputs=None,
    ):
        super().__init__(
            cache_dir=cache_dir,
            build_name=build_name,
            stats=stats,
            tensor_type=tensor_type,
            device_type=device_type,
            iterations=iterations,
            runtime=runtime,
            runtimes_supported=["trt"],
            runtime_version=TRT_VERSION,
            base_path=os.path.dirname(__file__),
            model=model,
            inputs=inputs,
            requires_docker=True,
        )

    def _setup(self) -> None:
        # Check if at least one NVIDIA GPU is available locally
        result = subprocess.run(
            ["nvidia-smi"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
            check=False,
        )

        if "NVIDIA" not in result.stdout or result.returncode == 1:
            msg = "No NVIDIA GPUs available on the local machine"
            raise exp.BenchmarkException(msg)

    def _execute(
        self,
        output_dir: str,
        onnx_file: str,
        outputs_file: str,
    ):
        power_readings = []
        average_power = 0

        # Start power measurement in a separate thread
        start_event = threading.Event()
        stop_event = threading.Event()
        power_thread = threading.Thread(
            target=measure_power, args=(start_event, stop_event, power_readings, 0.01)
        )

        # Add the GPU driver version to the stats file before execution
        gpu_driver_version = _get_nvidia_driver_version()
        self.stats.save_model_eval_stat("gpu_driver_version", gpu_driver_version)
        power_thread.start()

        run(
            output_dir=output_dir,
            onnx_file=onnx_file,
            outputs_file=outputs_file,
            errors_file=os.path.join(output_dir, "nvidia_error_log.txt"),
            iterations=self.iterations,
            start_event=start_event,
            stop_event=stop_event,
        )

        # Wait for power measurement to finish
        power_thread.join()

        # Calculate the average power consumption, average utilization, and peak power consumption
        average_power, average_utilization = average_power_and_utilization(
            power_readings
        )
        peak_power = (
            max([reading[1] for reading in power_readings]) if power_readings else None
        )

        # Load existing GPU performance data
        with open(outputs_file, "r", encoding="utf-8") as out_file:
            gpu_performance = json.load(out_file)

        # Add average power consumption, average utilization,
        # and peak power consumption to the dictionary
        gpu_performance["Average power consumption (W)"] = (
            round(average_power, 2) if average_power is not None else None
        )
        gpu_performance["Peak power consumption (W)"] = (
            round(peak_power, 2) if peak_power is not None else None
        )
        gpu_performance["Average GPU utilization (%)"] = (
            round(average_utilization, 2) if average_utilization is not None else None
        )

        # Save the updated GPU performance data
        with open(outputs_file, "w", encoding="utf-8") as out_file:
            json.dump(gpu_performance, out_file, ensure_ascii=False, indent=4)

    @property
    def mean_latency(self):
        return float(self._get_stat("Total Latency")["mean "].split(" ")[1])

    @property
    def throughput(self):
        return float(self._get_stat("Throughput").split(" ")[0])

    @property
    def device_name(self):
        return self._get_stat("Selected Device")



# labels: name::densenet121 author::timm task::Computer_Vision license::apache-2.0
import torch
import timm
from turnkeyml.parser import parse

torch.manual_seed(0)

# Parsing command-line arguments
batch_size, pretrained = parse(["batch_size", "pretrained"])

# Creating model and set it to evaluation mode
model = timm.create_model("densenet121", pretrained = pretrained)
model.eval()

# Creating inputs
input_size = model.default_cfg["input_size"]
batched_input_size = (batch_size,) + input_size
inputs = torch.rand(batched_input_size)

# Calling model
model(inputs)



from .runtime import TorchRT

implements = {
    "runtimes": {
        "torch-eager": {
            "build_required": False,
            "RuntimeClass": TorchRT,
            "supported_devices": {"x86"},
        },
        "torch-compiled": {
            "build_required": False,
            "RuntimeClass": TorchRT,
            "supported_devices": {"x86"},
        },
    }
}



import subprocess
import logging
import os
import sys
from typing import List, Optional, Dict

TIMEOUT = 900


class SubprocessError(Exception):
    pass


class CondaError(Exception):
    """
    Triggered when execution within the Conda environment goes wrong
    """


def run_subprocess(cmd):
    """Run a subprocess with the given command and log the output."""
    if isinstance(cmd, str):
        cmd_str = cmd
        shell_flag = True
    else:
        cmd_str = " ".join(cmd)
        shell_flag = False

    try:
        result = subprocess.run(
            cmd,
            check=True,
            capture_output=True,
            text=True,
            timeout=TIMEOUT,
            shell=shell_flag,
        )
        return result
    except subprocess.TimeoutExpired:
        logging.error(f"{cmd_str} timed out after {TIMEOUT} seconds")
        raise SubprocessError("TimeoutExpired")
    except subprocess.CalledProcessError as e:
        logging.error(
            f"Subprocess failed with command: {cmd_str} and error message: {e.stderr}"
        )
        raise SubprocessError("CalledProcessError")
    except (OSError, ValueError) as e:
        logging.error(
            f"Subprocess failed with command: {cmd_str} and error message: {str(e)}"
        )
        raise SubprocessError(str(e))


def get_python_path(conda_env_name):
    try:
        conda_path = os.getenv("CONDA_EXE")
        cmd = [
            conda_path,
            "run",
            "--name",
            conda_env_name,
            "python",
            "-c",
            "import sys; print(sys.executable)",
        ]
        result = subprocess.run(
            cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        python_path = result.stdout.decode().strip()

        return python_path
    except subprocess.CalledProcessError as e:
        raise EnvironmentError(
            f"An error occurred while getting Python path for {conda_env_name} environment"
            f"{e.stderr.decode()}"
        )


def logged_subprocess(
    cmd: List[str],
    cwd: str = os.getcwd(),
    env: Optional[Dict] = None,
    log_file_path: Optional[str] = None,
    log_to_std_streams: bool = True,
    log_to_file: bool = True,
) -> None:
    """
    This function calls a subprocess and sends the logs to either a file, stdout/stderr, or both.

    cmd             Command that will run o a sbprocess
    cwd             Working directory from where the subprocess should run
    env             Evironment to be used by the subprocess (useful for passing env vars)
    log_file_path   Where logs will be stored
    log_to_file     Whether or not to store the subprocess's stdout/stderr into a file
    log_to_std      Whether or not to print subprocess's stdout/stderr to the screen
    """
    if env is None:
        env = os.environ.copy()
    if log_to_file and log_file_path is None:
        raise ValueError("log_file_path must be set when log_to_file is True")

    log_stdout = ""
    log_stderr = ""
    try:
        proc = subprocess.run(
            cmd,
            check=True,
            env=env,
            capture_output=True,
            cwd=cwd,
        )
    except Exception as e:  # pylint: disable=broad-except
        log_stdout = e.stdout.decode("utf-8")  # pylint: disable=no-member
        log_stderr = e.stderr.decode("utf-8")  # pylint: disable=no-member
        raise CondaError(
            f"Exception {e} encountered, \n\nstdout was: "
            f"\n{log_stdout}\n\n and stderr was: \n{log_stderr}"
        )
    else:
        log_stdout = proc.stdout.decode("utf-8")
        log_stderr = proc.stderr.decode("utf-8")
    finally:
        if log_to_std_streams:
            # Print log to stdout
            # This might be useful when this subprocess is being logged externally
            print(log_stdout, file=sys.stdout)
            print(log_stderr, file=sys.stdout)
        if log_to_file:
            log = f"{log_stdout}\n{log_stderr}"
            with open(
                log_file_path,
                "w",
                encoding="utf-8",
            ) as f:
                f.write(log)



import sys
import os
import inspect
import importlib.util
import copy
import time
import shlex
import functools
import dataclasses
import traceback
import hashlib
from typing import Union, List, Dict, Tuple, Optional
from types import FrameType, TracebackType
from enum import Enum
import torch
import git
from turnkeyml.common import printing
import turnkeyml.common.build as build
import turnkeyml.common.exceptions as exp
from turnkeyml.build.stage import Sequence
import turnkeyml.analyze.status as status
import turnkeyml.analyze.util as util
import turnkeyml.common.tf_helpers as tf_helpers
import turnkeyml.common.labels as labels
from turnkeyml.build_api import build_model
import turnkeyml.common.filesystem as fs
import turnkeyml.run.devices as plugins


class Action(Enum):
    ANALYZE = "analyze"
    BUILD = "build"
    BENCHMARK = "benchmark"


@dataclasses.dataclass
class TracerArgs:
    input: str
    script_args: str
    device: str
    runtime: str
    iterations: int
    actions: List[Action]
    lean_cache: bool
    targets: List[str]
    max_depth: int
    onnx_opset: int
    cache_dir: str
    rebuild: str
    models_found: Dict[str, util.ModelInfo] = dataclasses.field(default_factory=dict)
    script_name: Optional[str] = None
    sequence: Optional[Sequence] = None
    rt_args: Optional[Dict] = None

    @functools.cached_property
    def labels(self) -> Dict[str, str]:
        # Load labels data from python scripts
        # This is not compatible with ONNX files, so we return
        # and empty dictionary in that case
        if self.input.endswith(".py"):
            return labels.load_from_file(self.input)
        else:
            return {}

    @functools.cached_property
    def torch_activations(self) -> List[str]:
        act = tf_helpers.get_classes(torch.nn.modules.activation)
        act += tf_helpers.get_transformers_activations()
        return act


def _store_traceback(invocation_info: util.UniqueInvocationInfo):
    """
    Store the traceback from an exception into invocation_info so that
    we can print it during the status update.
    """

    exc_type, exc_value, exc_traceback = sys.exc_info()
    invocation_info.traceback = traceback.format_exception(
        exc_type, exc_value, exc_traceback
    )

    # Remove line breaks and sequences of spaces from status message
    invocation_info.status_message = " ".join(invocation_info.status_message.split())


def set_status_on_exception(build_state: build.State, stats: fs.Stats):
    # We get `state` when the build tool succeeds, so we can use that to identify
    # whether the exception was thrown during build or benchmark
    if not build_state:
        stats.save_model_eval_stat(
            fs.Keys.BUILD_STATUS, build.FunctionStatus.ERROR.value
        )
    else:
        stats.save_model_eval_stat(
            fs.Keys.BENCHMARK_STATUS, build.FunctionStatus.ERROR.value
        )


def explore_invocation(
    model_inputs: dict,
    model_info: util.ModelInfo,
    invocation_info: util.UniqueInvocationInfo,
    tracer_args: TracerArgs,
) -> None:
    """
    Calls the turnkey function from within the model forward function
    """

    # Update status to "computing"
    invocation_info.status_message = "Computing..."
    invocation_info.status_message_color = printing.Colors.OKBLUE

    build_name = fs.get_build_name(
        tracer_args.script_name, tracer_args.labels, invocation_info.hash
    )
    status.update(tracer_args.models_found, build_name, tracer_args.cache_dir)

    # Organize the inputs to python model instances
    # Not necessary for ONNX models
    if model_info.model_type == build.ModelType.ONNX_FILE:
        inputs = model_inputs
    else:
        # Get a copy of the keyword arguments
        args, kwargs = model_inputs
        inputs = {}
        for k in kwargs.keys():
            if torch.is_tensor(kwargs[k]):
                inputs[k] = torch.tensor(kwargs[k].detach().numpy())
            else:
                inputs[k] = copy.deepcopy(kwargs[k])

        # Convert all positional arguments into keyword arguments
        if args != ():
            if model_info.model_type in [
                build.ModelType.PYTORCH,
                build.ModelType.PYTORCH_COMPILED,
            ]:
                forward_function = model_info.model.forward
            elif model_info.model_type == build.ModelType.KERAS:
                forward_function = model_info.model.call
            all_args = list(inspect.signature(forward_function).parameters.keys())
            for i in range(len(args)):
                if torch.is_tensor(args[i]):
                    inputs[all_args[i]] = torch.tensor(args[i].detach().numpy())
                else:
                    inputs[all_args[i]] = args[i]
        invocation_info.inputs = inputs

    # Create a build directory in the cache
    fs.make_build_dir(tracer_args.cache_dir, build_name)

    # If the user has not provided a specific runtime, select the runtime
    # based on the device provided.
    (
        selected_runtime,
        runtime_info,
        sequence_selected,
    ) = plugins.select_runtime_and_sequence(
        tracer_args.device,
        tracer_args.runtime,
        tracer_args.sequence,
    )

    if "status_stats" in runtime_info.keys():
        invocation_info.stats_keys = runtime_info["status_stats"]
    else:
        invocation_info.stats_keys = []

    # Create an ID for the build stats by combining the device and runtime.
    # We don't need more info in the evaluation_id because changes to build_model()
    # arguments (e.g., sequence) will trigger a rebuild, which is intended to replace the
    # build stats so long as the device and runtime have not changed.
    evaluation_id = f"{tracer_args.device}_{selected_runtime}"

    stats = fs.Stats(
        tracer_args.cache_dir,
        build_name,
        evaluation_id,
    )
    invocation_info.stats = stats

    # Stats that apply to the model, regardless of build
    stats.save_model_stat(
        fs.Keys.HASH,
        model_info.hash,
    )
    stats.save_model_stat(
        fs.Keys.MODEL_NAME,
        tracer_args.script_name,
    )
    stats.save_model_stat(
        fs.Keys.PARAMETERS,
        model_info.params,
    )
    if model_info.model_type != build.ModelType.ONNX_FILE:
        stats.save_model_stat(fs.Keys.CLASS, type(model_info.model).__name__)
    if fs.Keys.AUTHOR in tracer_args.labels:
        stats.save_model_stat(fs.Keys.AUTHOR, tracer_args.labels[fs.Keys.AUTHOR][0])
    if fs.Keys.TASK in tracer_args.labels:
        stats.save_model_stat(fs.Keys.TASK, tracer_args.labels[fs.Keys.TASK][0])

    # Save the system information used for this evaluation
    system_info = build.get_system_info()
    stats.save_model_stat(
        fs.Keys.SYSTEM_INFO,
        system_info,
    )

    # Save all of the lables in one place
    stats.save_model_stat(fs.Keys.LABELS, tracer_args.labels)

    # If the input script is a built-in TurnkeyML model, make a note of
    # which one
    if os.path.abspath(fs.MODELS_DIR) in os.path.abspath(tracer_args.input):
        try:
            # If this turnkey installation is in a git repo, use the
            # specific git hash
            git_repo = git.Repo(search_parent_directories=True)
            git_hash = git_repo.head.object.hexsha
        except git.exc.InvalidGitRepositoryError:
            # If we aren't in a git repo (e.g., PyPI package), point the user back to main
            git_hash = "main"

        relative_path = tracer_args.input.replace(
            fs.MODELS_DIR,
            f"https://github.com/onnx/turnkeyml/tree/{git_hash}/models",
        ).replace("\\", "/")
        stats.save_model_stat(fs.Keys.MODEL_SCRIPT, relative_path)

    # Build-specific stats
    stats.save_model_eval_stat(
        fs.Keys.DEVICE_TYPE,
        tracer_args.device,
    )
    stats.save_model_eval_stat(
        fs.Keys.RUNTIME,
        selected_runtime,
    )
    stats.save_model_eval_stat(
        fs.Keys.ITERATIONS,
        tracer_args.iterations,
    )

    if model_info.model_type == build.ModelType.PYTORCH_COMPILED:
        invocation_info.status_message = (
            "Skipping model compiled using torch.compile(). "
            "turnkey requires models to be in eager mode "
            "(regardless of what runtime you have selected)."
        )
        invocation_info.status_message_color = printing.Colors.WARNING

        return

    # Initialize build and benchmark status to "not started"
    stats.save_model_eval_stat(
        fs.Keys.BUILD_STATUS, build.FunctionStatus.NOT_STARTED.value
    )
    stats.save_model_eval_stat(
        fs.Keys.BENCHMARK_STATUS, build.FunctionStatus.NOT_STARTED.value
    )

    build_state = None
    perf = None
    try:
        # Run the build tool (if needed by the runtime)
        if runtime_info["build_required"]:
            # Indicate that the build is running. If the build fails for any reason,
            # we will try to catch the exception and note it in the stats.
            # If a concluded build still has a status of "running", this means
            # there was an uncaught exception.
            stats.save_model_eval_stat(
                fs.Keys.BUILD_STATUS, build.FunctionStatus.INCOMPLETE.value
            )

            build_state = build_model(
                model=model_info.model,
                inputs=inputs,
                evaluation_id=evaluation_id,
                build_name=build_name,
                cache_dir=tracer_args.cache_dir,
                rebuild=tracer_args.rebuild,
                sequence=sequence_selected,
                onnx_opset=tracer_args.onnx_opset,
                device=tracer_args.device,
            )

            stats.save_model_eval_stat(
                fs.Keys.BUILD_STATUS, build.FunctionStatus.SUCCESSFUL.value
            )

            model_to_benchmark = build_state.results[0]

            # Analyze the onnx file (if any) and save statistics
            util.analyze_onnx(
                build_name=build_name,
                cache_dir=tracer_args.cache_dir,
                stats=stats,
            )
        else:
            model_to_benchmark = model_info.model

        # Run the benchmark tool (if requested by the user)
        if Action.BENCHMARK in tracer_args.actions:
            if tracer_args.rt_args is None:
                rt_args_to_use = {}
            else:
                rt_args_to_use = tracer_args.rt_args

            stats.save_model_eval_stat(
                fs.Keys.BENCHMARK_STATUS, build.FunctionStatus.INCOMPLETE.value
            )

            model_handle = runtime_info["RuntimeClass"](
                cache_dir=tracer_args.cache_dir,
                build_name=build_name,
                stats=stats,
                iterations=tracer_args.iterations,
                model=model_to_benchmark,
                inputs=inputs,
                device_type=tracer_args.device,
                runtime=selected_runtime,
                **rt_args_to_use,
            )
            perf = model_handle.benchmark()

            for key, value in vars(perf).items():
                stats.save_model_eval_stat(
                    key=key,
                    value=value,
                )

            stats.save_model_eval_stat(
                fs.Keys.BENCHMARK_STATUS, build.FunctionStatus.SUCCESSFUL.value
            )

            invocation_info.status_message = "Model successfully benchmarked!"
            invocation_info.performance = perf
            invocation_info.status_message_color = printing.Colors.OKGREEN
        else:
            invocation_info.status_message = "Model successfully built!"
            invocation_info.status_message_color = printing.Colors.OKGREEN

    except exp.StageError as e:
        invocation_info.status_message = f"Build Error: {e}"
        invocation_info.status_message_color = printing.Colors.WARNING

        set_status_on_exception(build_state, stats)

        _store_traceback(invocation_info)

    except exp.SkipBuild:
        # SkipBuild is an exception that the build_model() API will raise
        # when it is skipping a previously-failed build when rebuild=never is set

        # NOTE: skipping a build should never update build or benchmark status

        invocation_info.status_message = (
            "Build intentionally skipped because rebuild=never"
        )
        invocation_info.status_message_color = printing.Colors.WARNING

    except exp.ArgError as e:
        # ArgError indicates that some argument to build_model() or BaseRT was
        # illegal. In that case we want to halt execution so that users can
        # fix their arguments.

        set_status_on_exception(build_state, stats)

        raise e

    except exp.Error as e:
        invocation_info.status_message = f"Error: {e}."
        invocation_info.status_message_color = printing.Colors.WARNING

        set_status_on_exception(build_state, stats)

        _store_traceback(invocation_info)

    # This broad exception is ok since enumerating all exceptions is
    # not possible, as the tested software continuously evolves.
    except Exception as e:  # pylint: disable=broad-except
        invocation_info.status_message = f"Unknown turnkey error: {e}"
        invocation_info.status_message_color = printing.Colors.WARNING

        set_status_on_exception(build_state, stats)

        _store_traceback(invocation_info)

    finally:
        # Ensure that stdout/stderr is not being forwarded before updating status
        util.stop_logger_forward()

        status.update(tracer_args.models_found, build_name, tracer_args.cache_dir)

        if tracer_args.lean_cache:
            printing.log_info("Removing build artifacts...")
            fs.clean_output_dir(tracer_args.cache_dir, build_name)


def get_model_hash(
    model: Union[torch.nn.Module, "tf.keras.Model", str], model_type: build.ModelType
):
    return build.hash_model(
        model, model_type, hash_params=model_type == build.ModelType.ONNX_FILE
    )[:8]


def get_invocation_hash(
    model_hash: str, parent_invocation_hash: str, args: Tuple, kwargs: Dict
) -> str:
    """
    Combines the model hash and the input shapes to create the invocation hash
    We also ensure that invocations that come from different parents have different hashes
    """

    # Merge positional and keyword args
    args = {"Positional Arg {}".format(i + 1): arg for i, arg in enumerate(args)}
    kwargs = {**kwargs, **args}

    # Get input shapes and types
    input_shapes, input_dtypes = build.get_shapes_and_dtypes(kwargs)

    hashable_content = (
        f"{model_hash}{parent_invocation_hash}{input_shapes}{input_dtypes}"
    )
    return hashlib.sha256(hashable_content.encode()).hexdigest()[:8], input_shapes


def store_model_info(
    model: Union[torch.nn.Module, "tf.keras.Model"],
    model_name: str,
    model_type: build.ModelType,
    frame: FrameType,
    event: str,
    tracer_args: TracerArgs,
    depth: int,
    parent_hash: str,
):
    # Getting the model hash is only possible after the first inference of Keras models
    model_hash = get_model_hash(model, model_type)

    # File where the model was found
    file = str(frame)[str(frame).find("file ") + 6 : str(frame).find("',")]

    # Line where the model was found
    line = frame.f_lineno if event == "return" else frame.f_lineno - 1

    # Keep track of all models details

    # If we have already found a model, don't add it to models_found again
    # We have to use both the model hash and the script name, since we don't
    # want to ignore a model if it was explicitly called in two different scripts
    identifier = f"{model_hash}_{tracer_args.script_name}"
    model_already_found = False
    for model_info in tracer_args.models_found.values():
        if identifier == f"{model_info.hash}_{model_info.script_name}":
            model_already_found = True

    if not model_already_found:
        build_model = Action.BUILD in tracer_args.actions
        tracer_args.models_found[model_hash] = util.ModelInfo(
            model=model,
            name=model_name,
            file=file,
            line=line,
            depth=depth,
            hash=model_hash,
            parent_hash=parent_hash,
            build_model=build_model,
            model_type=model_type,
            script_name=tracer_args.script_name,
        )


def explore_frame(
    frame,
    event,
    local_var_name,
    local_var,
    tracer_args: TracerArgs,
    depth: int = 0,
    parent_hash: Union[str, None] = None,
):
    """
    This function checks whether local_var is a torch or keras model.
    If it is, we will modify its forward function to know when it
    is called.
    """

    # Exit frame exploration if Python is shutting down
    if not bool(sys.modules):
        return

    # Skip all variables that are not a subclass of torch.nn.Module/tf.keras.Model
    # Note: try block used since dead weakreferences fail when checking subclass
    try:
        if issubclass(type(local_var), torch.nn.Module):
            if type(local_var) in tracer_args.torch_activations:
                return
            if "dynamo_ctx" in local_var.__dict__:
                model_type = build.ModelType.PYTORCH_COMPILED
            else:
                model_type = build.ModelType.PYTORCH
        elif tf_helpers.is_keras_subclass(type(local_var)):
            model_type = build.ModelType.KERAS
        else:
            return
    except AttributeError:
        return

    # Skip self variable and variable names commonly used by child models
    if (
        local_var_name == "self"
        or local_var_name == "instance"
        or local_var_name == "child"
        or local_var_name == "layer"
        or local_var_name == "module"
    ):
        return

    # Check if we are inside of a subclass of torch.nn.Module or tf.keras.model
    inside_class = False
    inside_nn_subclass = False
    if "self" in frame.f_locals:
        self_var = frame.f_locals["self"]
        inside_class = type(self_var)
        inside_nn_subclass = issubclass(
            inside_class, torch.nn.Module
        ) or tf_helpers.is_keras_subclass(inside_class)

    if not inside_nn_subclass:
        if hasattr(local_var, "forward_instrumented"):
            # A previously-found model might have been compiled
            # Update that information if needed
            if model_type == build.ModelType.PYTORCH_COMPILED:
                tracer_args.models_found[
                    local_var.turnkey_hash
                ].model_type = build.ModelType.PYTORCH_COMPILED
            return

        if model_type == build.ModelType.PYTORCH:
            # Avoid instrumenting models before they have been fully loaded
            if util.count_parameters(local_var, model_type) == 0:
                return

            # Mark this model as instrumented
            local_var.forward_instrumented = True

            # Create a copy of the old forward function
            old_forward = local_var.forward

            # Recursively look for sub-models within the found model
            # This is only possible on Pytorch, since each layer of a torch.nn.module
            # is also a torch.nn.module.
            model_hash = get_model_hash(local_var, model_type)
            local_var.turnkey_hash = model_hash
            if depth < tracer_args.max_depth:
                recursive_search(
                    frame, event, local_var, depth, model_hash, tracer_args
                )

            # We can keep track of Pytorch models even before they are executed
            store_model_info(
                local_var,
                local_var_name,
                model_type,
                frame,
                event,
                tracer_args,
                depth,
                parent_hash,
            )
        elif model_type == build.ModelType.KERAS:
            # Mark this model as instrumented
            local_var.forward_instrumented = True

            # Create a copy of the old forward function
            old_forward = local_var.call

            # Raise exception if user tries to use max_depth!=0 for a keras model
            if tracer_args.max_depth != 0:
                raise exp.Error("max_depth is not supported for Keras models")
        local_var.old_forward = old_forward

        def forward_spy(*args, **kwargs):
            tracer = sys.getprofile()
            if tracer is not None:
                # Turn tracing off while the model is being executed for speed
                sys.setprofile(None)
            elif depth == 0:
                # If this model is being executed and the tracing is already off
                # we are calling a module within a parent module. We only run
                # on child models if the user has explicitly asked us to
                # do so by setting the max_depth flag.
                return old_forward(*args, **kwargs)

            # We can only keep track of keras models once they have been executed
            if model_type == build.ModelType.KERAS:
                store_model_info(
                    local_var,
                    local_var_name,
                    model_type,
                    frame,
                    event,
                    tracer_args,
                    depth,
                    parent_hash,
                )

            # Get parent invocation hash
            parent_invocation_hash = None
            if parent_hash:
                parent_invocation_hash = tracer_args.models_found[
                    parent_hash
                ].last_unique_invocation_executed

            model_hash = get_model_hash(local_var, model_type)
            invocation_hash, input_shapes = get_invocation_hash(
                model_hash, parent_invocation_hash, args, kwargs
            )
            model_info = tracer_args.models_found[model_hash]

            if invocation_hash not in model_info.unique_invocations:
                model_info.unique_invocations[
                    invocation_hash
                ] = util.UniqueInvocationInfo(
                    hash=invocation_hash,
                    is_target=invocation_hash in tracer_args.targets
                    or len(tracer_args.targets) == 0,
                    input_shapes=input_shapes,
                    parent_hash=parent_invocation_hash,
                )
            model_info.last_unique_invocation_executed = invocation_hash

            # Keep track of execution time
            start_time = time.time()
            outputs = old_forward(*args, **kwargs)
            end_time = time.time()

            invocation_info = model_info.unique_invocations[invocation_hash]
            invocation_info.exec_time = (
                invocation_info.exec_time + end_time - start_time
            )
            invocation_info.executed = invocation_info.executed + 1

            # Call explore_invocation() if this is the first time the model is being executed
            # and this model has been selected by the user
            if (
                invocation_info.executed == 1
                and invocation_info.is_target
                and (model_info.build_model)
            ):
                explore_invocation(
                    model_inputs=[args, kwargs],
                    model_info=model_info,
                    invocation_info=invocation_info,
                    tracer_args=tracer_args,
                )
                # Ensure that explore_invocation() doesn't interfere with our execution count
                model_info.executed = 1

            build_name = fs.get_build_name(
                tracer_args.script_name, tracer_args.labels, invocation_info.hash
            )
            status.update(tracer_args.models_found, build_name, tracer_args.cache_dir)

            # Turn tracing on again after computing the outputs
            sys.setprofile(tracer)

            return outputs

        # The inspect module offers the ability to actually copy the signature of the wrapped
        # function. This allows other functions to see the correct parameters instead of the
        # enigmatic *args, **kwargs. This is especially important for Keras, since it heavily
        # relies on inspections to the call function.
        forward_spy.__signature__ = inspect.signature(old_forward)

        # Use modified forward/call function
        if model_type == build.ModelType.PYTORCH:
            local_var.forward = forward_spy
        elif model_type == build.ModelType.KERAS:
            local_var.call = forward_spy


def tracefunc(
    frame: FrameType, event: str, _, tracer_args: TracerArgs
) -> TracebackType:
    """
    This function is used to trace the program as it runs in order
    to keep track of all all instantiated models.
    This function is passed to sys.setprofile() as a callback function.
    It receives three arguments:
        frame (the stack frame from the code being run),
        event (a string naming the type of notification), and
        arg (an event-specific value)

    """

    # Create a copy of f_locals.keys() to avoid errors due to dict changing
    local_names = list(frame.f_locals.keys())

    # Loop over all local variables to check if new models can be found
    for local_var_name in local_names:
        explore_frame(
            frame,
            event,
            local_var_name,
            frame.f_locals[local_var_name],
            tracer_args=tracer_args,
            depth=0,
        )

    return tracefunc


def recursive_search(
    frame: FrameType,
    event: str,
    model: Union[torch.nn.Module, "tf.keras.Model"],
    depth: int,
    parent_hash: Union[str, None],
    tracer_args: TracerArgs,
):
    """
    Recursively check for submodels within found models
    """
    element_names = list(dict(model.named_modules()).keys())[1:]
    for element_name in element_names:
        if hasattr(model, element_name):
            element = getattr(model, element_name)
            if issubclass(type(element), torch.nn.Module):
                explore_frame(
                    frame,
                    event,
                    element_name,
                    element,
                    tracer_args,
                    depth=depth + 1,
                    parent_hash=parent_hash,
                )


@dataclasses.dataclass
class HelpfulHandler:
    # Type of exception to handle
    exc_type: Exception
    # Do not print any traceback after this message is encountered
    traceback_stop_msg: str
    # Message to print that gives context to the traceback
    helpful_msg: str


class AnalysisException(Exception):
    pass


class HelpfulExceptions:
    """
    Catch certain exceptions, defined by `HelpfulHandler`s, and print a more helpful
    error message and traceback than what would ordinarily be printed out. This is
    useful to avoid showing the user a giant traceback that goes all the way through
    our profiling code.
    """

    def __init__(self, exceptions_to_handle: List[HelpfulHandler]):
        self.excs = exceptions_to_handle

    def __enter__(self):
        pass

    def __exit__(self, exc_type, _exc_value, exc_tb):
        for exc_handler in self.excs:
            if exc_type == exc_handler.exc_type:
                # Search the full traceback for the traceback_stop_msg
                tb = traceback.format_tb(exc_tb)

                # This default value of offending_line makes it so we will print
                # the entire traceback if we can't find the traceback_stop_msg
                offending_line = -2
                for i, line in enumerate(tb):
                    if exc_handler.traceback_stop_msg in line:
                        offending_line = i

                # Eliminate the lines of traceback before and after the traceback_stop_msg
                # Typically, the lines that follow will be related to our profiling
                # code and not helpful to the user

                # Start the helpful_traceback after line 3, since the first 3 lines are related
                # to our profiler
                start_line = 3
                helpful_traceback = "\n".join(tb[start_line : offending_line + 1])

                # sys.tracebacklimit = 0 prevents the unwanted traceback from printing
                # when we raise our AnalysisException
                sys.tracebacklimit = 0
                raise AnalysisException(
                    f"{exc_handler.helpful_msg}\n\nTraceback: \n\n: {helpful_traceback}"
                )


def evaluate_script(tracer_args: TracerArgs) -> Dict[str, util.ModelInfo]:
    tracer_args.script_name = fs.clean_file_name(tracer_args.input)

    # Get a pointer to the script's python module
    spec = importlib.util.spec_from_file_location("__main__", tracer_args.input)
    module = importlib.util.module_from_spec(spec)

    # Overwriting argv to import input script using "input-args"
    if tracer_args.script_args is None:
        tracer_args.script_args = []
    else:
        tracer_args.script_args = shlex.split(tracer_args.script_args)
    sys.argv = [tracer_args.input] + tracer_args.script_args
    sys.path.append(os.getcwd())

    # Create a tracer object that bundles a callback function with some args
    tracer = functools.partial(tracefunc, tracer_args=tracer_args)

    # Enabling analysis via setprofile
    sys.setprofile(tracer)

    # Import input script. Each executed frame of the input script will
    # trigger the tracefunc() callback function (defined above)
    with HelpfulExceptions(
        [
            HelpfulHandler(
                torch.jit.frontend.NotSupportedError,
                "torch.jit.script(",
                "torch.jit.script() is not supported by turnkey CLI and benchmark_files() API, "
                "however torch.jit.script() is being called in your script."
                "You can try passing your model instance into the build_model() API instead. ",
            )
        ]
    ):
        spec.loader.exec_module(module)

    # Stop profiling when we're done executing the module
    sys.setprofile(None)

    return tracer_args.models_found



# labels: test_group::turnkey name::unet author::torch_hub task::Computer_Vision license::bsd-3-clause
"""https://pytorch.org/hub/mateuszbuda_brain-segmentation-pytorch_unet/"""
from turnkeyml.parser import parse
import torch

# Parsing command-line arguments
batch_size, height, num_channels, width = parse(
    ["batch_size", "height", "num_channels", "width"]
)


# Model and input configurations
model = torch.hub.load(
    "mateuszbuda/brain-segmentation-pytorch",
    "unet",
    in_channels=3,
    out_channels=1,
    init_features=32,
    pretrained=True,
)

inputs = {"x": torch.ones(batch_size, num_channels, height, width, dtype=torch.float)}


# Call model
model(**inputs)



"""
The following script is used to get the latency and outputs of a given run on the x86 CPUs.
"""
# pylint: disable = no-name-in-module
# pylint: disable = import-error
import os
import subprocess
import json
from statistics import mean
import platform
import turnkeyml.run.plugin_helpers as plugin_helpers

ORT_VERSION = "1.15.1"

BATCHSIZE = 1


def create_conda_env(conda_env_name: str):
    """Create a Conda environment with the given name and install requirements."""
    conda_path = os.getenv("CONDA_EXE")
    if conda_path is None:
        raise EnvironmentError(
            "CONDA_EXE environment variable not set."
            "Make sure Conda is properly installed."
        )

    # Normalize the path for Windows
    if platform.system() == "Windows":
        conda_path = os.path.normpath(conda_path)

    env_path = os.path.join(
        os.path.dirname(os.path.dirname(conda_path)), "envs", conda_env_name
    )

    # Only create the environment if it does not already exist
    if not os.path.exists(env_path):
        plugin_helpers.run_subprocess(
            [
                conda_path,
                "create",
                "--name",
                conda_env_name,
                "python=3.8",
                "-y",
            ]
        )

    # Using conda run to execute pip install within the environment
    setup_cmd = [
        conda_path,
        "run",
        "--name",
        conda_env_name,
        "pip",
        "install",
        f"onnxruntime=={ORT_VERSION}",
    ]
    plugin_helpers.run_subprocess(setup_cmd)


def execute_benchmark(
    onnx_file: str,
    outputs_file: str,
    output_dir: str,
    conda_env_name: str,
    iterations: int,
):
    """Execute the benchmark script and retrieve the output."""

    python_in_env = plugin_helpers.get_python_path(conda_env_name)
    iterations_file = os.path.join(output_dir, "per_iteration_latency.json")
    benchmarking_log_file = os.path.join(output_dir, "ort_benchmarking_log.txt")

    cmd = [
        python_in_env,
        os.path.join(output_dir, "within_conda.py"),
        "--onnx-file",
        onnx_file,
        "--iterations",
        str(iterations),
        "--iterations-file",
        iterations_file,
    ]

    # Execute command and log stdout/stderr
    plugin_helpers.logged_subprocess(
        cmd=cmd,
        cwd=os.path.dirname(output_dir),
        log_to_std_streams=False,
        log_to_file=True,
        log_file_path=benchmarking_log_file,
    )

    # Parse per-iteration performance results and save aggregated results to a json file
    if os.path.exists(iterations_file):
        with open(iterations_file, "r", encoding="utf-8") as f:
            per_iteration_latency = json.load(f)
    else:
        raise ValueError(
            f"Execution of command {cmd} failed, see {benchmarking_log_file}"
        )

    cpu_performance = get_cpu_specs()
    cpu_performance["OnnxRuntime Version"] = str(ORT_VERSION)
    cpu_performance["Mean Latency(ms)"] = str(mean(per_iteration_latency) * 1000)
    cpu_performance["Throughput"] = str(BATCHSIZE / mean(per_iteration_latency))
    cpu_performance["Min Latency(ms)"] = str(min(per_iteration_latency) * 1000)
    cpu_performance["Max Latency(ms)"] = str(max(per_iteration_latency) * 1000)

    with open(outputs_file, "w", encoding="utf-8") as out_file:
        json.dump(cpu_performance, out_file, ensure_ascii=False, indent=4)


def get_cpu_specs() -> dict:
    # Define a common field mapping for both Windows and Linux
    field_mapping = {
        "Architecture": "CPU Architecture",
        "Manufacturer": "CPU Vendor",
        "MaxClockSpeed": "CPU Max Frequency (MHz)",
        "Name": "CPU Name",
        "NumberOfCores": "CPU Core Count",
        "Model name": "CPU Name",  # Additional mapping for Linux
        "CPU MHz": "CPU Max Frequency (MHz)",  # Additional mapping for Linux
        "CPU(s)": "CPU Core Count",  # Additional mapping for Linux
        "Vendor ID": "CPU Vendor",
    }

    # Check the operating system and define the command accordingly
    if platform.system() == "Windows":
        cpu_info_command = (
            "wmic CPU get Architecture,Manufacturer,MaxClockSpeed,"
            "Name,NumberOfCores /format:list"
        )

        cpu_info = subprocess.Popen(
            cpu_info_command, stdout=subprocess.PIPE, shell=True
        )
        separator = "="
    else:
        cpu_info_command = "lscpu"
        cpu_info = subprocess.Popen(cpu_info_command.split(), stdout=subprocess.PIPE)
        separator = ":"

    cpu_info_output, _ = cpu_info.communicate()
    if not cpu_info_output:
        raise EnvironmentError(
            f"Could not get CPU info using '{cpu_info_command.split()[0]}'. "
            "Please make sure this tool is correctly installed on your system before continuing."
        )

    decoded_info = (
        cpu_info_output.decode()
        .strip()
        .split("\r\n" if platform.system() == "Windows" else "\n")
    )

    # Initialize an empty dictionary to hold the CPU specifications
    cpu_spec = {}
    for line in decoded_info:
        key, value = line.split(separator, 1)
        # Get the corresponding key from the field mapping
        key = field_mapping.get(key.strip())
        if key:
            # Add the key and value to the CPU specifications dictionary
            cpu_spec[key] = value.strip()

    return cpu_spec



# labels: name::linear author::selftest test_group::selftest
import torch

torch.manual_seed(0)


class LinearTestModel(torch.nn.Module):
    def __init__(self, input_features, output_features):
        super(LinearTestModel, self).__init__()
        self.fc = torch.nn.Linear(input_features, output_features)

    def forward(self, x):
        output = self.fc(x)
        return output


input_features = 10
output_features = 10

# Model and input configurations
model = LinearTestModel(input_features, output_features)
inputs = {"x": torch.rand(input_features)}


# Call model
model(**inputs)



# labels: test_group::turnkey name::mealv2_mobilenetv3_small_100 author::torch_hub task::Computer_Vision license::bsd-3-clause
"""
https://github.com/pytorch/hub/blob/master/pytorch_vision_meal_v2.md
"""

from turnkeyml.parser import parse
import torch

torch.manual_seed(0)

# Parsing command-line arguments
batch_size, num_channels, width, height = parse(
    ["batch_size", "num_channels", "width", "height"]
)


# Model and input configurations
model = torch.hub.load(
    "szq0214/MEAL-V2",
    "meal_v2",
    "mealv2_mobilenetv3_small_100",
)
model.eval()
inputs = {"x": torch.ones([batch_size, num_channels, width, height])}


# Call model
model(**inputs)



"""
The following script is used to get the latency and outputs of a given run on the NVIDIA GPU.
"""
# pylint: disable = no-name-in-module
# pylint: disable = import-error
import subprocess
import logging
import json
import time
import threading

# Set a 15 minutes timeout for all docker commands
TIMEOUT = 900

TRT_VERSION = "23.09-py3"


def run(
    output_dir: str,
    onnx_file: str,
    outputs_file: str,
    errors_file: str,
    iterations: int,
    start_event: threading.Event,
    stop_event: threading.Event,
):
    # Latest docker image can be found here:
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/tensorrt/tags
    # TurnkeyML maintainers to keep this up to date with the latest version of release container
    latest_trt_docker = f"nvcr.io/nvidia/tensorrt:{TRT_VERSION}"
    docker_name = f"tensorrt{TRT_VERSION}"

    # docker run args:
    # "--gpus all" - use all gpus available
    # "-v <path>" -  mount the home dir to access the model inside the docker
    # "-itd" - start the docker in interactive mode in the background
    # "--rm" - remove the container automatically upon stopping
    docker_run_args = f"--gpus all -v {output_dir}:/app  -itd --rm"

    # docker exec args:
    # "--onnx=<path>" - path to the onnx model in the mounted file
    # "--fp16" - enable execution in fp16 mode on tensorrt
    # "--iterations=<int>" - number of iterations to run on tensorrt
    docker_exec_args = f"trtexec --onnx={onnx_file} --fp16 --iterations={iterations}"

    run_command = (
        f"docker run --name {docker_name} {docker_run_args} {latest_trt_docker}"
    )
    exec_command = f"docker exec {docker_name} {docker_exec_args}"
    stop_command = f"docker stop {docker_name}"

    # Run TensorRT docker in interactive model
    run_docker = subprocess.Popen(run_command.split(), stdout=subprocess.PIPE)
    try:
        run_docker.communicate(timeout=TIMEOUT)
    except subprocess.TimeoutExpired:
        logging.error(f"{run_command} timed out!")
    except subprocess.CalledProcessError as e:
        logging.error(f"{run_command} failed with error code {e.returncode}: {e}")

    # Start power measurement
    start_event.set()

    # Execute the onnx model user trtexec inside the container
    run_trtexec = subprocess.Popen(exec_command.split(), stdout=subprocess.PIPE)
    try:
        trtexec_output, trtexec_error = run_trtexec.communicate(timeout=TIMEOUT)
    except subprocess.TimeoutExpired:
        logging.error(f"{exec_command} timed out!")
    except subprocess.CalledProcessError as e:
        logging.error(f"{exec_command}  failed with error code {e.returncode}: {e}")

    # Stop power measurement
    stop_event.set()

    # Stop the container
    stop_docker = subprocess.Popen(stop_command.split(), stdout=subprocess.PIPE)
    try:
        stop_docker.communicate(timeout=TIMEOUT)
    except subprocess.TimeoutExpired:
        logging.error(f"{stop_command} timed out!")
    except subprocess.CalledProcessError as e:
        logging.error(f"{stop_command} failed with error code {e.returncode}: {e}")

    output = str(trtexec_output)
    error = str(trtexec_error)

    decoded_output = bytes(output, "utf-8").decode("unicode_escape")
    decoded_error = bytes(error, "utf-8").decode("unicode_escape")

    field_mapping = {
        "Selected Device": "Selected Device",
        "TensorRT version": "TensorRT version",
        "H2D Latency:": "Host to Device Latency",
        "GPU Compute Time:": "GPU Compute Time:",
        "D2H Latency:": "Device to Host Latency",
        "Throughput": "Throughput",
        "Latency:": "Total Latency",
    }

    def format_field(line: str):
        all_metrics = line.split(":")[-1].strip()
        if "," not in all_metrics:
            return all_metrics
        metrics = {
            k: v for k, v in (m.strip().split("=") for m in all_metrics.split(","))
        }
        return metrics

    gpu_performance = {}
    for line in decoded_output.splitlines():
        for field, key in field_mapping.items():
            if field in line:
                gpu_performance[key] = format_field(line)
                break
    with open(outputs_file, "w", encoding="utf-8") as out_file:
        json.dump(gpu_performance, out_file, ensure_ascii=False, indent=4)

    with open(errors_file, "w", encoding="utf-8") as e:
        e.writelines(decoded_error)


class NvidiaSmiError(Exception):
    """An exception class for errors related to nvidia-smi."""


# In the average_power_and_utilization function, we eliminate the values below
# the idle_threshold at the beginning and end of the list because we want to
# exclude the idle periods before and after the actual workload execution. By
# doing this, we can get a more accurate representation of the average power
# consumption and GPU utilization during the execution of the workload.
# We set the Idle threshold to 3% utilization based on heuristics.

IDLE_THRESHOLD = 3


def average_power_and_utilization(power_readings):
    if not power_readings:
        return 0, 0

    # Remove readings below the threshold from the beginning of the list
    while power_readings and power_readings[0][0] <= IDLE_THRESHOLD:
        power_readings.pop(0)

    # Remove readings below the threshold from the end of the list
    while power_readings and power_readings[-1][0] <= IDLE_THRESHOLD:
        power_readings.pop()

    if not power_readings:
        return 0, 0

    average_power = sum(power_draw for _, power_draw in power_readings) / len(
        power_readings
    )
    average_utilization = sum(utilization for utilization, _ in power_readings) / len(
        power_readings
    )

    return average_power, average_utilization


def get_gpu_power_and_utilization():
    try:
        query = "nvidia-smi --query-gpu=utilization.gpu,power.draw --format=csv,noheader,nounits"
        output = subprocess.check_output(query.split())
        output = output.decode("utf-8").strip()

        utilization, power_draw = output.split(", ")
        return int(utilization), float(power_draw)

    except subprocess.CalledProcessError as e:
        raise NvidiaSmiError(f"nvidia-smi failed with return code {e.returncode}")


def measure_power(start_event, stop_event, power_readings, sample_rate=0.01):
    start_event.wait()  # Wait for the start event to be set

    while not stop_event.is_set():
        gpu_utilization, power_draw = get_gpu_power_and_utilization()
        power_readings.append((gpu_utilization, power_draw))
        time.sleep(sample_rate)



from typing import Optional
from typing import List, Dict, Tuple
import turnkeyml.run.onnxrt as onnxrt
import turnkeyml.run.tensorrt as tensorrt
import turnkeyml.run.torchrt as torchrt
import turnkeyml.common.plugins as plugins
from turnkeyml.build.stage import Sequence
import turnkeyml.build.sequences as sequences
import turnkeyml.common.exceptions as exp


def supported_devices_list(data: Dict, parent_key: str = "") -> List:
    """Recursive function to generate all device::part::config pairs"""
    result = []
    for key, value in data.items():
        current_key = f"{parent_key}::{key}" if parent_key else key
        result.append(current_key)
        if isinstance(value, dict):
            result.extend(supported_devices_list(value, current_key))
        elif isinstance(value, list):
            for item in value:
                result.append(f"{current_key}::{item}")
    return result


discovered_plugins = plugins.discover()

DEFAULT_RUNTIME = 0

# Note: order matters here. We append the discovered_plugins after builtin so
# that the default runtime for each device will come from a builtin, whenever
# available.
builtin_runtimes = [onnxrt, tensorrt, torchrt]
plugin_modules = builtin_runtimes + list(discovered_plugins.values())

SUPPORTED_RUNTIMES = {}

for module in plugin_modules:
    if "runtimes" in module.implements.keys():
        for runtime_name, runtime_info in module.implements["runtimes"].items():
            if runtime_name in SUPPORTED_RUNTIMES:
                raise ValueError(
                    f"Your turnkeyml installation has two runtimes named '{runtime_name}' "
                    "installed. You must uninstall one of your plugins that includes "
                    f"{runtime_name}. Your imported runtime plugins are: {SUPPORTED_RUNTIMES}\n"
                    f"This error was thrown while trying to import {module}"
                )
            if isinstance(runtime_info["supported_devices"], set):
                runtime_info["supported_devices"] = {
                    item: {} for item in runtime_info["supported_devices"]
                }
            SUPPORTED_RUNTIMES[runtime_name] = runtime_info

# Get the list of supported devices by checking which devices each runtime supports
SUPPORTED_DEVICES = []
for runtime_info in SUPPORTED_RUNTIMES.values():
    SUPPORTED_DEVICES.extend(supported_devices_list(runtime_info["supported_devices"]))
SUPPORTED_DEVICES = list(set(SUPPORTED_DEVICES))

# Organizing the supported devices that are in the "long form".
# Those are not the elements used for setting the defaults.
# This is useful for nicely showing the list of supported devices as part of the help menu.
SUPPORTED_DEVICES.sort()

# Create a map of devices to the runtimes that support them
DEVICE_RUNTIME_MAP = {key: [] for key in SUPPORTED_DEVICES}
for device in SUPPORTED_DEVICES:
    for runtime_name, runtime_info in SUPPORTED_RUNTIMES.items():
        if device in supported_devices_list(runtime_info["supported_devices"]):
            DEVICE_RUNTIME_MAP[device].append(runtime_name)


def apply_default_runtime(device: str, runtime: Optional[str] = None):
    if runtime is None:
        return DEVICE_RUNTIME_MAP[device][DEFAULT_RUNTIME]
    else:
        return runtime


def _check_suggestion(value: str):
    return (
        f"You may need to check the spelling of '{value}', install a "
        "plugin, or update the turnkeyml package."
    )


def select_runtime_and_sequence(
    device: str, runtime: Optional[str], sequence: Optional[Sequence]
) -> Tuple[str, str, Sequence]:
    selected_runtime = apply_default_runtime(device, runtime)

    # Validate device and runtime selections
    if device not in SUPPORTED_DEVICES:
        raise exp.ArgError(
            f"Device argument '{device}' is not one of the available "
            f"supported devices {SUPPORTED_DEVICES}\n"
            f"{_check_suggestion(device)}"
        )
    if selected_runtime not in DEVICE_RUNTIME_MAP[device]:
        raise exp.ArgError(
            f"Runtime argument '{selected_runtime}' is not one of the available "
            f"runtimes supported for device '{device}': {DEVICE_RUNTIME_MAP[device]}\n"
            f"{_check_suggestion(selected_runtime)}"
        )

    # Get the plugin module for the selected runtime
    runtime_info = SUPPORTED_RUNTIMES[selected_runtime]

    # Perform a build, if necessary
    if runtime_info["build_required"]:
        # Get the build sequence that will be used for the model
        if sequence is None:
            # Automatically choose a Sequence based on what the runtime expects
            sequence_selected = runtime_info["default_sequence"]
        else:
            # User-specified Sequence
            if isinstance(sequence, str):
                # Sequence is defined by a plugin
                if sequence in sequences.SUPPORTED_SEQUENCES.keys():
                    sequence_selected = sequences.SUPPORTED_SEQUENCES[sequence]
                else:
                    raise ValueError(
                        f"Sequence argument {sequence} is not one of the "
                        "available sequences installed: "
                        f"{sequences.SUPPORTED_SEQUENCES.keys()} \n"
                        f"{_check_suggestion(sequence)}"
                    )

            elif isinstance(sequence, Sequence):
                # Sequence is a user-defined instance of Sequence
                sequence_selected = sequence

    else:
        # Sequence is only needed for builds
        sequence_selected = None

    return selected_runtime, runtime_info, sequence_selected



import turnkeyml.build.export as export
import turnkeyml.build.stage as stage
import turnkeyml.common.plugins as plugins

optimize_fp16 = stage.Sequence(
    "optimize_fp16",
    "Optimized FP16 ONNX file",
    [
        export.ExportPlaceholder(),
        export.OptimizeOnnxModel(),
        export.ConvertOnnxToFp16(),
    ],
    enable_model_validation=True,
)

optimize_fp32 = stage.Sequence(
    "optimize_fp32",
    "Optimized FP32 ONNX File",
    [
        export.ExportPlaceholder(),
        export.OptimizeOnnxModel(),
    ],
    enable_model_validation=True,
)

onnx_fp32 = stage.Sequence(
    "onnx_fp32",
    "Base Sequence",
    [
        export.ExportPlaceholder(),
    ],
    enable_model_validation=True,
)

# Plugin interface for sequences
discovered_plugins = plugins.discover()

# Populated supported sequences dict with builtin sequences
SUPPORTED_SEQUENCES = {
    "optimize-fp16": optimize_fp16,
    "optimize-fp32": optimize_fp32,
    "onnx-fp32": onnx_fp32,
}

# Add sequences from plugins to supported sequences dict
for module in discovered_plugins.values():
    if "sequences" in module.implements.keys():
        for seq_name, seq_info in module.implements["sequences"].items():
            if seq_name in SUPPORTED_SEQUENCES:
                raise ValueError(
                    f"Your turnkeyml installation has two sequences named '{seq_name}' "
                    "installed. You must uninstall one of your plugins that includes "
                    f"{seq_name}. Your imported sequence plugins are: {SUPPORTED_SEQUENCES}\n"
                    f"This error was thrown while trying to import {module}"
                )

            SUPPORTED_SEQUENCES[seq_name] = seq_info["sequence_instance"]



import os
from typing import Optional, List, Dict, Any
import turnkeyml.build.ignition as ignition
import turnkeyml.build.stage as stage
import turnkeyml.common.printing as printing
import turnkeyml.common.build as build
import turnkeyml.common.filesystem as filesystem


def build_model(
    model: build.UnionValidModelInstanceTypes = None,
    inputs: Optional[Dict[str, Any]] = None,
    build_name: Optional[str] = None,
    evaluation_id: Optional[str] = "build",
    cache_dir: str = filesystem.DEFAULT_CACHE_DIR,
    monitor: Optional[bool] = None,
    rebuild: Optional[str] = None,
    sequence: Optional[List[stage.Stage]] = None,
    onnx_opset: Optional[int] = None,
    device: Optional[str] = None,
) -> build.State:
    """Use build a model instance into an optimized ONNX file.

    Args:
        model: Model to be mapped to an optimized ONNX file, which can be a PyTorch
            model instance, Keras model instance, Hummingbird model instance,
            or a path to an ONNX file.
        inputs: Example inputs to the user's model. The ONNX file will be
            built to handle inputs with the same static shape only.
        build_name: Unique name for the model that will be
            used to store the ONNX file and build state on disk. Defaults to the
            name of the file that calls build_model().
        evaluation_id: Unique name for evaluation statistics that should persist across multiple
            builds of the same model.
        cache_dir: Directory to use as the cache for this build. Output files
            from this build will be stored at cache_dir/build_name/
            Defaults to the current working directory, but we recommend setting it to
            an absolute path of your choosing.
        monitor: Display a monitor on the command line that
            tracks the progress of this function as it builds the ONNX file.
        rebuild: determines whether to rebuild or load a cached build. Options:
            - "if_needed" (default): overwrite invalid cached builds with a rebuild
            - "always": overwrite valid cached builds with a rebuild
            - "never": load cached builds without checking validity, with no guarantee
                of functionality or correctness
            - None: Falls back to default
        sequence: Override the default sequence of build stages. Power
            users only.
        onnx_opset: ONNX opset to use during ONNX export.
        device: Specific device target to take into account during the build sequence.
            Use the format "device_family", "device_family::part", or
            "device_family::part::configuration" to refer to a family of devices,
            part within a family, or configuration of a part model, respectively.

        More information is available in the Tools User Guide:
            https://github.com/onnx/turnkeyml/blob/main/docs/tools_user_guide.md
    """

    # Allow monitor to be globally disabled by an environment variable
    if monitor is None:
        if os.environ.get("TURNKEY_BUILD_MONITOR") == "False":
            monitor_setting = False
        else:
            monitor_setting = True
    else:
        monitor_setting = monitor

    # Support "~" in the cache_dir argument
    parsed_cache_dir = os.path.expanduser(cache_dir)

    # Validate and lock in the config (user arguments that
    # configure the build) that will be used by the rest of the toolchain
    config = ignition.lock_config(
        model=model,
        build_name=build_name,
        sequence=sequence,
        onnx_opset=onnx_opset,
        device=device,
    )

    # Analyze the user's model argument and lock in the model, inputs,
    # and sequence that will be used by the rest of the toolchain
    (
        model_locked,
        inputs_locked,
        sequence_locked,
        model_type,
    ) = ignition.model_intake(
        model,
        inputs,
        sequence,
    )

    # Get the state of the model from the cache if a valid build is available
    state = ignition.load_or_make_state(
        config=config,
        evaluation_id=evaluation_id,
        cache_dir=parsed_cache_dir,
        rebuild=rebuild or build.DEFAULT_REBUILD_POLICY,
        model_type=model_type,
        monitor=monitor_setting,
        model=model_locked,
        inputs=inputs_locked,
    )

    # Return a cached build if possible, otherwise prepare the model State for
    # a build
    if state.build_status == build.FunctionStatus.SUCCESSFUL:
        # Successful builds can be loaded from cache and returned with
        # no additional steps
        additional_msg = " (build_name auto-selected)" if config.auto_name else ""
        printing.log_success(
            f' Build "{config.build_name}"{additional_msg} found in cache. Loading it!',
        )

        return state

    sequence_locked.show_monitor(config, state.monitor)
    state = sequence_locked.launch(state)

    printing.log_success(
        f"\n    Saved to **{build.output_dir(state.cache_dir, config.build_name)}**"
    )

    return state



import argparse
import re
import math
import json
import time
import numpy as np
import onnxruntime as ort


def run_ort_profile(
    onnx_file_path: str,
    iterations_file: str,
    iterations: int,
):
    # Run the provided onnx model using onnxruntime and measure average latency

    per_iteration_latency = []
    sess_options = ort.SessionOptions()
    sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
    onnx_session = ort.InferenceSession(onnx_file_path, sess_options)
    sess_input = onnx_session.get_inputs()
    input_feed = dummy_inputs(sess_input)
    output_name = onnx_session.get_outputs()[0].name

    for _ in range(iterations):
        start = time.perf_counter()
        onnx_session.run([output_name], input_feed)
        end = time.perf_counter()
        iteration_latency = end - start
        per_iteration_latency.append(iteration_latency)

    with open(iterations_file, "w", encoding="utf-8") as out_file:
        json.dump(per_iteration_latency, out_file, ensure_ascii=False, indent=4)


def dummy_inputs(sess_input) -> dict:
    # Generate dummy inputs of the expected shape and type for the input model
    input_stats = []
    for _idx, input_ in enumerate(range(len(sess_input))):
        input_name = sess_input[input_].name
        input_shape = sess_input[input_].shape

        # TODO: Use onnx update_inputs_outputs_dims to automatically freeze models
        for dim in input_shape:
            if isinstance(dim, str) is True or math.isnan(dim) is True:
                raise AssertionError(
                    "Error: Model has dynamic inputs. Freeze the graph and try again"
                )

        input_type = sess_input[input_].type
        input_stats.append([input_name, input_shape, input_type])

    input_feed = {}
    for stat in input_stats:
        dtype_str = re.search(r"\((.*)\)", stat[2])
        assert dtype_str is not None
        datatype = dtype_ort2str(dtype_str.group(1))
        input_feed[stat[0]] = np.random.rand(*stat[1]).astype(datatype)
    return input_feed


def dtype_ort2str(dtype_str: str):
    if dtype_str == "float16":
        datatype = "float16"
    elif dtype_str == "float":
        datatype = "float32"
    elif dtype_str == "double":
        datatype = "float64"
    elif dtype_str == "long":
        datatype = "int64"
    else:
        datatype = dtype_str
    return datatype


if __name__ == "__main__":
    # Parse Inputs
    parser = argparse.ArgumentParser(description="Execute models using onnxruntime")
    parser.add_argument(
        "--onnx-file",
        required=True,
        help="Path where the ONNX file is located",
    )
    parser.add_argument(
        "--iterations-file",
        required=True,
        help="File in which to place the per-iteration execution timings",
    )
    parser.add_argument(
        "--iterations",
        required=True,
        type=int,
        help="Number of times to execute the received onnx model",
    )
    args = parser.parse_args()

    run_ort_profile(
        onnx_file_path=args.onnx_file,
        iterations_file=args.iterations_file,
        iterations=args.iterations,
    )



import os
from typing import Dict, Any
from statistics import mean
import time
from packaging import version
import torch
import numpy as np
from turnkeyml.run.basert import BaseRT
from turnkeyml.common.performance import MeasuredPerformance
from turnkeyml.run.onnxrt.execute import get_cpu_specs
import turnkeyml.build.ignition as ignition
import turnkeyml.common.build as build
import turnkeyml.common.exceptions as exp
from turnkeyml.common.filesystem import Stats


class TorchRT(BaseRT):
    def __init__(
        self,
        cache_dir: str,
        build_name: str,
        stats: Stats,
        device_type: str,
        runtime: str,
        iterations: int,
        model: torch.nn.Module,
        inputs: Dict[str, Any],
        tensor_type=np.array,
    ):
        self.throughput_ips = None
        self.mean_latency_ms = None

        super().__init__(
            cache_dir=cache_dir,
            build_name=build_name,
            stats=stats,
            device_type=device_type,
            runtime=runtime,
            iterations=iterations,
            runtimes_supported=["torch-eager", "torch-compiled"],
            runtime_version=str(torch.__version__),
            base_path=os.path.dirname(__file__),
            tensor_type=tensor_type,
            model=model,
            inputs=inputs,
        )

    def _setup(self) -> None:
        # Ensure we have the correct model type
        model_type = ignition.identify_model_type(self.model)
        if model_type != build.ModelType.PYTORCH:
            raise exp.IntakeError(
                f"Only Pytorch models are valid when runtime is {self.runtime}"
            )

        # Compile the model
        if self.runtime == "torch-compiled":
            # First ensure we have the required version of Pytorch
            clean_torch_version = self.runtime_version.split("+")[0]
            if version.parse(clean_torch_version) < version.parse("2.0.0"):
                exp.BenchmarkException(
                    (
                        f"{self.runtime} can only be used with Pytorch 2.0.0 or above. "
                        f"However, version {self.runtime_version} was found."
                    )
                )

            self.model = torch.compile(self.model)

    def benchmark(self) -> MeasuredPerformance:
        per_iteration_latency = [0] * self.iterations
        for idx in range(self.iterations):
            start_time = time.perf_counter()
            self.model(**self.inputs)
            end_time = time.perf_counter()
            per_iteration_latency[idx] = end_time - start_time

        # Calculate perf from per_iteration_latency
        self.mean_latency_ms = mean(per_iteration_latency) * 1000
        self.throughput_ips = float(
            1 / (np.sum(per_iteration_latency) / self.iterations)
        )

        return MeasuredPerformance(
            mean_latency=self.mean_latency,
            throughput=self.throughput,
            device=self.device_name,
            device_type=self.device_type,
            runtime=self.runtime,
            runtime_version=self.runtime_version,
            build_name=self.build_name,
        )

    @property
    def mean_latency(self) -> float:
        if self.mean_latency_ms is not None:
            return self.mean_latency_ms
        else:
            raise exp.BenchmarkException(
                "Queried mean latency before self.benchmark() was called"
            )

    @property
    def throughput(self) -> float:
        if self.throughput_ips is not None:
            return self.throughput_ips
        else:
            raise exp.BenchmarkException(
                "Queried throughput before self.benchmark() was called"
            )

    @property
    def device_name(self) -> str:
        return get_cpu_specs()["CPU Name"]



import os
import onnx
import onnx.tools.update_model_dims
import numpy as np

import turnkeyml.common.build as build
import turnkeyml.build.stage as stage
import turnkeyml.build.export as export
import turnkeyml.common.exceptions as exp
import turnkeyml.common.filesystem as fs

try:
    # An initial selection of Hummingbird-supported models.
    from sklearn.ensemble import ExtraTreesClassifier
    from sklearn.ensemble import GradientBoostingClassifier
    from sklearn.ensemble import IsolationForest
    from sklearn.ensemble import RandomForestClassifier
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import SGDClassifier
    from sklearn.naive_bayes import BernoulliNB
    from sklearn.naive_bayes import GaussianNB
    from sklearn.naive_bayes import MultinomialNB
    from sklearn.neighbors import KNeighborsClassifier
    from sklearn.neural_network import MLPClassifier
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import StandardScaler
    from sklearn.svm import LinearSVC
    from sklearn.tree import DecisionTreeClassifier

    sklearn_available = True
except ImportError as e:
    sklearn_available = False

try:
    from xgboost import XGBClassifier
    from xgboost import XGBRegressor

    xgboost_available = True
except ImportError as e:
    xgboost_available = False

try:
    from lightgbm import LGBMClassifier
    from lightgbm import LGBMRegressor

    lightgbm_available = True
except ImportError as e:
    lightgbm_available = False


def is_supported_sklearn_model(model) -> bool:
    return (
        isinstance(model, ExtraTreesClassifier)
        or isinstance(model, GradientBoostingClassifier)
        or isinstance(model, IsolationForest)
        or isinstance(model, RandomForestClassifier)
        or isinstance(model, RandomForestRegressor)
        or isinstance(model, SGDClassifier)
        or isinstance(model, BernoulliNB)
        or isinstance(model, GaussianNB)
        or isinstance(model, MultinomialNB)
        or isinstance(model, KNeighborsClassifier)
        or isinstance(model, MLPClassifier)
        or isinstance(model, Pipeline)
        or isinstance(model, StandardScaler)
        or isinstance(model, LinearSVC)
        or isinstance(model, DecisionTreeClassifier)
    )


def is_supported_xgboost_model(model) -> bool:
    return isinstance(model, XGBClassifier) or isinstance(model, XGBRegressor)


def is_supported_lightgbm_model(model) -> bool:
    return isinstance(model, LGBMClassifier) or isinstance(model, LGBMRegressor)


def is_supported_model(model) -> bool:
    return (
        (sklearn_available and is_supported_sklearn_model(model))
        or (xgboost_available and is_supported_xgboost_model(model))
        or (lightgbm_available and is_supported_lightgbm_model(model))
    )


class ConvertHummingbirdModel(stage.Stage):
    """
    Stage that takes an SKLearn, XGBoost, or LightGBM model instance, in state.model, and
    converts it to an ONNX file via Hummingbird.

    Expected inputs:
     - state.model is an SKLearn, XGBoost, or LightGBM model object
     - state.inputs is a dict of the form {"input_0": <data>} where <data>
        is a numpy array as would be provided, e.g., to sklearn's predict method.

    Outputs:
     - A *-base.onnx file that implements state.model given state.inputs
    """

    def __init__(self):
        super().__init__(
            unique_name="hummingbird_conversion",
            monitor_message="Converting model to ONNX with Hummingbird",
        )

    def fire(self, state: build.State):
        # TODO: Temporarily inlined to avoid warning message in hummingbird-ml<=0.46.
        import hummingbird.ml  # pylint: disable=import-error
        from hummingbird.ml.exceptions import (  # pylint: disable=import-error
            ConstantError,
            MissingConverter,
            MissingBackend,
        )

        if not is_supported_model(state.model):
            msg = f"""
            The current stage (ConvertHummingbirdModel) is only compatible with
            certain scikit-learn, xgboost, and lightgbm models, however the stage
            received an unsupported model of type {type(state.model)}.

            Support scikit-learn models:
              - sklearn.ensemble.ExtraTreesClassifier
              - sklearn.ensemble.GradientBoostingClassifier
              - sklearn.ensemble.IsolationForest
              - sklearn.ensemble.RandomForestClassifier
              - sklearn.ensemble.RandomForestRegressor
              - sklearn.linear_model.SGDClassifier
              - sklearn.naive_bayes.BernoulliNB
              - sklearn.naive_bayes.GaussianNB
              - sklearn.naive_bayes.MultinomialNB
              - sklearn.neighbors.KNeighborsClassifier
              - sklearn.neural_network.MLPClassifier
              - sklearn.pipeline.Pipeline
              - sklearn.preprocessing.StandardScaler
              - sklearn.svm.LinearSVC
              - sklearn.tree.DecisionTreeClassifier

            Supported xgboost models:
              - xgboost.XGBClassifier
              - xgboost.XGBRegressor

            Supported lightgbm models:
              - lightgbm.LGBMClassifier
              - lightgbm.LGBMRegressor
            """
            raise exp.StageError(msg)

        # TODO: By default the strategy will be chosen wih Hummingbird's logic.
        # Ideally, this would also be a parameter.
        tree_implementation_strategy = "gemm"  # or "tree_trav" or "perf_tree_trav"

        inputs = state.inputs
        if inputs is None:
            raise exp.StageError(
                "Hummingbird conversion requires inputs to be provided,"
                " however `inputs` is None."
            )
        test_X = inputs["input_0"]
        batch_size = test_X.shape[0]
        if test_X.dtype == np.float64:
            raise exp.StageError(
                "Fitting a model with float64 inputs can cause issues"
                " with conversion and compilation. This can be corrected by changing"
                " code like model.fit(X, y) to model.fit(X.astype(numpy.float32), y)."
            )

        extra_config = {
            "onnx_target_opset": state.config.onnx_opset,
            "tree_implementation": tree_implementation_strategy,
            "batch_size": batch_size,
        }

        try:
            onnx_model = hummingbird.ml.convert(
                state.model, "onnx", test_X, extra_config=extra_config
            ).model
        except (
            RuntimeError,
            IndexError,
            ValueError,
            ConstantError,
            MissingConverter,
            MissingBackend,
        ) as e:
            raise exp.StageError(f"Hummingbird conversion failed with error: {e}")

        input_dims = {
            "input_0": [
                batch_size,
                onnx_model.graph.input[0].type.tensor_type.shape.dim[1].dim_value,
            ]
        }
        if len(onnx_model.graph.output) > 1:
            output_dims = {
                "variable": [batch_size],
                onnx_model.graph.output[1].name: [
                    batch_size,
                    onnx_model.graph.output[1].type.tensor_type.shape.dim[1].dim_value,
                ],
            }
        else:
            output_dims = {"variable": [batch_size]}

        # Concretize symbolic shape parameter
        onnx_model = onnx.tools.update_model_dims.update_inputs_outputs_dims(
            onnx_model, input_dims, output_dims
        )

        # Save output node names
        state.expected_output_names = export.get_output_names(onnx_model)

        output_path = export.base_onnx_file(state)
        os.makedirs(export.onnx_dir(state))
        onnx.save(onnx_model, output_path)

        np.save(state.original_inputs_file, state.inputs)

        state.intermediate_results = [output_path]
        stats = fs.Stats(state.cache_dir, state.config.build_name, state.evaluation_id)
        stats.save_model_eval_stat(
            fs.Keys.ONNX_FILE,
            output_path,
        )

        return state



from dataclasses import dataclass
from typing import Optional, Union, Dict
import turnkeyml.common.printing as printing
import turnkeyml.common.exceptions as exp


class Device:
    def __init__(self, selected_device: str, rt_supported_devices: Optional[Dict] = None):
        self.family: str
        self.part: Optional[str] = None
        self.config: Optional[str] = None

        # Unpack selected_device
        values = selected_device.split("::")
        if len(values) > 3:
            raise exp.ArgError(
                f"Recieved a device argument that has more than 3 members: {selected_device}. "
                "Please format device arguments as either `family` (1 member), "
                "`family::part` (2 members), or `family::part::config` (3 members)."
            )

        # Set family, part, and config straight away if rt_supported_devices is not provided
        if rt_supported_devices is None:
            if len(values)>0:
                self.family = values[0]
            if len(values)>1:
                self.part = values[1]
            if len(values)>2:
                self.config = values[2]
            return

        # If rt_supported_devices is provided we:
        # (1) Check whether the received family/part/config are in rt_supported_devices
        # (2) Set part/config to the default values (if needed)
        # Note: The default values are set according to the order in which elements were
        #       added to rt_supported_devices. The first element is set as the default.
        #       This feature relies on dictionaries being ordered (Python 3.6+).

        # Set family
        if values[0] in rt_supported_devices:
            self.family = values[0]
        else:
            raise exp.ArgError(
                f"Family {values[0]} is not supported by this device. "
                f"Supported families are: {rt_supported_devices.keys()}"
            )

        # Set part (assign to default if no needed)
        if len(values) > 1:
            if values[1] in rt_supported_devices[self.family]:
                self.part = values[1]
            elif len(rt_supported_devices[self.family]) == 0:
                raise exp.ArgError(
                    f"Device family {self.family} supports no parts."
                )
            else:
                error_msg = f"Part {values[1]} is not supported by this device family."
                if len(rt_supported_devices[self.family]) > 0:
                    error_msg += f" Supported parts are: {rt_supported_devices[self.family]}"
                raise exp.ArgError(error_msg)
        elif rt_supported_devices[self.family]:
            self.part = next(iter(rt_supported_devices[self.family]))

        # Set config (assign to default if no needed)
        if len(values) > 2:
            supported_configs = rt_supported_devices[self.family][self.part]
            if values[2] in supported_configs:
                self.config = values[2]
            else:
                error_msg = f"Config {values[2]} is not supported by this device family and part."
                if len(supported_configs) > 0:
                    error_msg += f"Supported configs are: {supported_configs}"
                raise exp.ArgError(error_msg)
        elif self.part is not None:
            if rt_supported_devices[self.family][self.part]:
                self.config = rt_supported_devices[self.family][self.part][0]

    def __str__(self) -> str:
        result = self.family

        if self.part:
            result = result + "::" + self.part

            if self.config:
                result = result + "::" + self.config

        return result


@dataclass
class MeasuredPerformance:
    throughput: float
    mean_latency: float
    device: str
    runtime: str
    runtime_version: str
    device_type: Union[str, Device]
    build_name: str
    throughput_units: str = "inferences per second (IPS)"
    latency_units: str = "milliseconds (ms)"

    def print(self):
        printing.log_info(
            f"\nPerformance of build {self.build_name} on {self.device} "
            f"({self.runtime} v{self.runtime_version}) is:"
        )
        print(f"\tMean Latency: {self.mean_latency:.3f} {self.latency_units}")
        print(f"\tThroughput: {self.throughput:.1f} {self.throughput_units}")
        print()

    def __post_init__(self):
        if isinstance(self.device_type, Device):
            self.device_type = str(self.device_type)



import argparse
import os
import sys
import copy
from difflib import get_close_matches
import turnkeyml.common.build as build
import turnkeyml.common.exceptions as exceptions
import turnkeyml.common.filesystem as filesystem
import turnkeyml.cli.report as report
import turnkeyml.cli.parser_helpers as parser_helpers
from turnkeyml.files_api import benchmark_files
from turnkeyml.version import __version__ as turnkey_version
from turnkeyml.run.devices import SUPPORTED_DEVICES, SUPPORTED_RUNTIMES
from turnkeyml.build.sequences import SUPPORTED_SEQUENCES
from turnkeyml.cli.spawn import DEFAULT_TIMEOUT_SECONDS


class MyParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write(f"error: {message}\n\n")
        sys.stderr.write(f"Run '{self.prog} --help' for more information\n\n")
        self.print_usage(sys.stderr)
        sys.exit(2)


def print_version(_):
    """
    Print the package version number
    """
    print(turnkey_version)


def print_stats(args):
    state_path = build.state_file(args.cache_dir, args.build_name)
    filesystem.print_yaml_file(state_path, "build state")

    filesystem.print_yaml_file(
        filesystem.Stats(args.cache_dir, args.build_name).file, "stats"
    )


def benchmark_command(args):
    """
    Map the argparse args into benchmark_files() arguments

    Assumes the following rules:
    -   All args passed to a "benchmark" command should be forwarded to the benchmark_files()
        API, except as explicitly handled below.
    -   The "dest" names of all CLI args must exactly match the names of the corresponding API arg
    """

    api_args = copy.deepcopy(vars(args))

    # Remove the function ID because it was only used to get us into this method
    api_args.pop("func")

    # Decode CLI arguments before calling the API
    api_args["rt_args"] = parser_helpers.decode_args(api_args["rt_args"])

    benchmark_files(**api_args)


def main():
    """
    Parses arguments passed by user and forwards them into a
    command function
    """

    parser = MyParser(
        description="TurnkeyML benchmarking command line interface",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # We use sub-parsers to keep the help info neatly organized for each command
    # Sub-parses also allow us to set command-specific help on options like --cache-dir
    # that are used in multiple commands

    subparsers = parser.add_subparsers(
        title="command",
        help="Choose one of the following commands:",
        metavar="COMMAND",
        required=True,
    )

    #######################################
    # Parser for the "benchmark" command
    #######################################

    def check_extension(choices, file_name, error_func):
        _, extension = os.path.splitext(file_name.split("::")[0])
        if extension[1:].lower() not in choices:
            error_func(
                f"input_files must end with .py, .onnx, or .txt (got '{file_name}')\n"
            )
        return file_name

    benchmark_parser = subparsers.add_parser(
        "benchmark",
        help="Benchmark the performance of one or more models",
        description="Analyze, build, and then benchmark the model(s) within input file(s).",
    )
    benchmark_parser.set_defaults(func=benchmark_command)

    benchmark_parser.add_argument(
        "input_files",
        nargs="+",
        help="One or more script (.py), ONNX (.onnx), or input list (.txt) files to be benchmarked",
        type=lambda file: check_extension(
            ("py", "onnx", "txt"), file, benchmark_parser.error
        ),
    )

    toolchain_select_group = benchmark_parser.add_argument_group(
        "Select which phase(s) of the toolchain to run "
        "(default is to run analyze, build, and benchmark)"
    )

    toolchain_select_group.add_argument(
        "-a",
        "--analyze-only",
        dest="analyze_only",
        help="Stop this command after the analyze phase",
        action="store_true",
    )

    toolchain_select_group.add_argument(
        "-b",
        "--build-only",
        dest="build_only",
        help="Stop this command after the analyze and build phases",
        action="store_true",
    )

    analyze_group = benchmark_parser.add_argument_group(
        "Options that specifically apply to the `analyze` phase of the toolflow"
    )

    analyze_group.add_argument(
        "--labels",
        dest="labels",
        help="Only benchmark the scripts that have the provided labels",
        nargs="*",
        default=[],
    )

    analyze_group.add_argument(
        "--script-args",
        dest="script_args",
        type=str,
        help="Arguments to pass into the target script(s)",
    )

    analyze_group.add_argument(
        "--max-depth",
        dest="max_depth",
        type=int,
        default=0,
        help="Maximum depth to analyze within the model structure of the target script(s)",
    )

    both_build_benchmark_group = benchmark_parser.add_argument_group(
        "Options that apply to both the `build` and `benchmark` phases of the toolflow"
    )

    benchmark_default_device = "x86"
    both_build_benchmark_group.add_argument(
        "--device",
        choices=SUPPORTED_DEVICES,
        dest="device",
        help="Type of hardware device to be used for the benchmark "
        f'(defaults to "{benchmark_default_device}")',
        required=False,
        default=benchmark_default_device,
    )

    both_build_benchmark_group.add_argument(
        "--runtime",
        choices=SUPPORTED_RUNTIMES.keys(),
        dest="runtime",
        help="Software runtime that will be used to collect the benchmark. "
        "Must be compatible with the selected device. "
        "Automatically selects a sequence if `--sequence` is not used."
        "If this argument is not set, the default runtime of the selected device will be used.",
        required=False,
        default=None,
    )

    both_build_benchmark_group.add_argument(
        "-d",
        "--cache-dir",
        dest="cache_dir",
        help="Build cache directory where the resulting build directories will "
        f"be stored (defaults to {filesystem.DEFAULT_CACHE_DIR})",
        required=False,
        default=filesystem.DEFAULT_CACHE_DIR,
    )

    both_build_benchmark_group.add_argument(
        "--lean-cache",
        dest="lean_cache",
        help="Delete all build artifacts except for log files when the command completes",
        action="store_true",
    )

    build_group = benchmark_parser.add_argument_group(
        "Options that apply specifically to the `build` phase of the toolflow"
    )

    build_group.add_argument(
        "--sequence",
        choices=SUPPORTED_SEQUENCES.keys(),
        dest="sequence",
        help="Name of a build sequence that will define the model-to-model transformations, "
        "used to build the models. Each runtime has a default sequence that it uses.",
        required=False,
        default=None,
    )

    build_group.add_argument(
        "--rebuild",
        choices=build.REBUILD_OPTIONS,
        dest="rebuild",
        help=f"Sets the cache rebuild policy (defaults to {build.DEFAULT_REBUILD_POLICY})",
        required=False,
        default=build.DEFAULT_REBUILD_POLICY,
    )

    build_group.add_argument(
        "--onnx-opset",
        dest="onnx_opset",
        type=int,
        default=None,
        help=f"ONNX opset used when creating ONNX files (default={build.DEFAULT_ONNX_OPSET}). "
        "Not applicable when input model is already a .onnx file.",
    )

    benchmark_group = benchmark_parser.add_argument_group(
        "Options that apply specifically to the `benchmark` phase of the toolflow"
    )

    benchmark_group.add_argument(
        "--iterations",
        dest="iterations",
        type=int,
        default=100,
        help="Number of execution iterations of the model to capture\
              the benchmarking performance (e.g., mean latency)",
    )

    benchmark_group.add_argument(
        "--rt-args",
        dest="rt_args",
        type=str,
        nargs="*",
        help="Optional arguments provided to the runtime being used",
    )

    all_toolflows_group = benchmark_parser.add_argument_group(
        "Options that apply to all toolflows"
    )

    slurm_or_processes_group = all_toolflows_group.add_mutually_exclusive_group()

    slurm_or_processes_group.add_argument(
        "--use-slurm",
        dest="use_slurm",
        help="Execute on Slurm instead of using local compute resources",
        action="store_true",
    )

    slurm_or_processes_group.add_argument(
        "--process-isolation",
        dest="process_isolation",
        help="Isolate evaluating each input into a separate process",
        action="store_true",
    )

    all_toolflows_group.add_argument(
        "--timeout",
        type=int,
        default=None,
        help="Build timeout, in seconds, after which a build will be canceled "
        f"(default={DEFAULT_TIMEOUT_SECONDS}). Only "
        "applies when --process-isolation or --use-slurm is also used.",
    )

    #######################################
    # Subparser for the "cache" command
    #######################################

    cache_parser = subparsers.add_parser(
        "cache",
        help="Commands for managing the build cache",
    )

    cache_subparsers = cache_parser.add_subparsers(
        title="cache",
        help="Commands for managing the build cache",
        required=True,
        dest="cache_cmd",
    )

    #######################################
    # Parser for the "cache report" command
    #######################################

    report_parser = cache_subparsers.add_parser(
        "report", help="Generate reports in CSV format"
    )
    report_parser.set_defaults(func=report.summary_spreadsheets)

    report_parser.add_argument(
        "-d",
        "--cache-dir",
        dest="cache_dirs",
        help=(
            "One or more build cache directories to generate the report "
            f"(defaults to {filesystem.DEFAULT_CACHE_DIR})"
        ),
        default=[filesystem.DEFAULT_CACHE_DIR],
        nargs="*",
    )

    report_parser.add_argument(
        "-r",
        "--report-dir",
        dest="report_dir",
        help="Path to folder where report will be saved (defaults to current working directory)",
        required=False,
        default=os.getcwd(),
    )

    #######################################
    # Parser for the "cache list" command
    #######################################

    list_parser = cache_subparsers.add_parser(
        "list", help="List all builds in a target cache"
    )
    list_parser.set_defaults(func=filesystem.print_available_builds)

    list_parser.add_argument(
        "-d",
        "--cache-dir",
        dest="cache_dir",
        help="The builds in this build cache directory will printed to the terminal "
        f" (defaults to {filesystem.DEFAULT_CACHE_DIR})",
        required=False,
        default=filesystem.DEFAULT_CACHE_DIR,
    )

    #######################################
    # Parser for the "cache stats" command
    #######################################

    stats_parser = cache_subparsers.add_parser(
        "stats", help="Print stats about a build in a target cache"
    )
    stats_parser.set_defaults(func=print_stats)

    stats_parser.add_argument(
        "-d",
        "--cache-dir",
        dest="cache_dir",
        help="The stats of a build in this build cache directory will printed to the terminal "
        f" (defaults to {filesystem.DEFAULT_CACHE_DIR})",
        required=False,
        default=filesystem.DEFAULT_CACHE_DIR,
    )

    stats_parser.add_argument(
        "build_name",
        help="Name of the specific build whose stats are to be printed, within the cache directory",
    )

    #######################################
    # Parser for the "cache delete" command
    #######################################

    delete_parser = cache_subparsers.add_parser(
        "delete", help="Delete one or more builds in a build cache"
    )
    delete_parser.set_defaults(func=filesystem.delete_builds)

    delete_parser.add_argument(
        "-d",
        "--cache-dir",
        dest="cache_dir",
        help="Search path for builds " f"(defaults to {filesystem.DEFAULT_CACHE_DIR})",
        required=False,
        default=filesystem.DEFAULT_CACHE_DIR,
    )

    delete_group = delete_parser.add_mutually_exclusive_group(required=True)

    delete_group.add_argument(
        "build_name",
        nargs="?",
        help="Name of the specific build to be deleted, within the cache directory",
    )

    delete_group.add_argument(
        "--all",
        dest="delete_all",
        help="Delete all builds in the cache directory",
        action="store_true",
    )

    #######################################
    # Parser for the "cache clean" command
    #######################################

    clean_parser = cache_subparsers.add_parser(
        "clean",
        help="Remove the build artifacts from one or more builds in a build cache",
    )
    clean_parser.set_defaults(func=filesystem.clean_builds)

    clean_parser.add_argument(
        "-d",
        "--cache-dir",
        dest="cache_dir",
        help="Search path for builds " f"(defaults to {filesystem.DEFAULT_CACHE_DIR})",
        required=False,
        default=filesystem.DEFAULT_CACHE_DIR,
    )

    clean_group = clean_parser.add_mutually_exclusive_group(required=True)

    clean_group.add_argument(
        "build_name",
        nargs="?",
        help="Name of the specific build to be cleaned, within the cache directory",
    )

    clean_group.add_argument(
        "--all",
        dest="clean_all",
        help="Clean all builds in the cache directory",
        action="store_true",
    )

    #######################################
    # Parser for the "cache location" command
    #######################################

    cache_location_parser = cache_subparsers.add_parser(
        "location",
        help="Print the location of the default build cache directory",
    )
    cache_location_parser.set_defaults(func=filesystem.print_cache_dir)

    #######################################
    # Subparser for the "models" command
    #######################################

    models_parser = subparsers.add_parser(
        "models",
        help="Commands for managing the models",
    )

    models_subparsers = models_parser.add_subparsers(
        title="models",
        help="Commands for managing the models",
        required=True,
        dest="models_cmd",
    )

    models_location_parser = models_subparsers.add_parser(
        "location",
        help="Print the location of the models directory",
    )
    models_location_parser.set_defaults(func=filesystem.print_models_dir)

    models_location_parser.add_argument(
        "--quiet",
        dest="verbose",
        help="Command output will only include the directory path",
        required=False,
        action="store_false",
    )

    #######################################
    # Parser for the "version" command
    #######################################

    version_parser = subparsers.add_parser(
        "version",
        help="Print the package version number",
    )
    version_parser.set_defaults(func=print_version)

    #######################################
    # Execute the command
    #######################################

    # The default behavior of this CLI is to run the build command
    # on a target script. If the user doesn't provide a command,
    # we alter argv to insert the command for them.

    # Special characters that indicate a string is a filename, not a command
    file_chars = [".", "/", "\\", "*"]

    if len(sys.argv) > 1:
        first_arg = sys.argv[1]
        if first_arg not in subparsers.choices.keys() and "-h" not in first_arg:
            if any(char_to_check in first_arg for char_to_check in file_chars):
                # User has provided a file as the first positional arg
                sys.argv.insert(1, "benchmark")
            else:
                # User has provided a command as the first positional arg
                # Check how close we are from each of the valid options
                # NOTE: if we are not close to a valid option, we will let
                # argparse detect and raise the error
                valid_options = list(subparsers.choices.keys())
                close_matches = get_close_matches(first_arg, valid_options)

                if close_matches:
                    raise exceptions.ArgError(
                        f"Unexpected command `turnkey {first_arg}`. "
                        f"Did you mean `turnkey {close_matches[0]}`?"
                    )

    args = parser.parse_args()

    args.func(args)


if __name__ == "__main__":
    main()



# labels: name::mt5_base author::transformers task::Generative_AI license::apache-2.0
from turnkeyml.parser import parse
from transformers import MT5Model, AutoConfig
import torch

torch.manual_seed(0)

# Parsing command-line arguments
pretrained, batch_size, max_seq_length = parse(
    ["pretrained", "batch_size", "max_seq_length"]
)

# Model and input configurations
if pretrained:
    model = MT5Model.from_pretrained("google/mt5-base")
else:
    config = AutoConfig.from_pretrained("google/mt5-base")
    model = MT5Model(config)


inputs = {
    "input_ids": torch.ones(batch_size, max_seq_length, dtype=torch.long),
    "decoder_input_ids": torch.ones(batch_size, max_seq_length, dtype=torch.long),
    "attention_mask": torch.ones(batch_size, max_seq_length, dtype=torch.float),
}


# Call model
model(**inputs)



import os
import shutil
import glob
import pathlib
from typing import Dict, List, Optional
import importlib.util
import yaml
import turnkeyml.common.printing as printing
import turnkeyml.common.build as build
import turnkeyml.common.exceptions as exp
from turnkeyml.common import labels

# Allow an environment variable to override the default
# location for the build cache
if os.environ.get("TURNKEY_CACHE_DIR"):
    DEFAULT_CACHE_DIR = os.path.expanduser(os.environ.get("TURNKEY_CACHE_DIR"))
else:
    DEFAULT_CACHE_DIR = os.path.expanduser("~/.cache/turnkey")

if " " in DEFAULT_CACHE_DIR:
    raise ValueError(
        f"Your turnkey cache directory is set to '{DEFAULT_CACHE_DIR}', "
        "however the space in that path will cause problems. Please set a cache directory "
        "that has no spaces by setting the TURNKEY_CACHE_DIR environment variable."
    )

if DEFAULT_CACHE_DIR.endswith("\\"):
    raise ValueError(
        f"Your turnkey cache directory is set to '{DEFAULT_CACHE_DIR}', "
        "however the trailing backslash (\\) in that path will cause problems. "
        "Please set a cache directory "
        "that has no trailing backslash by setting the TURNKEY_CACHE_DIR environment variable."
    )

CACHE_MARKER = ".turnkeycache"
BUILD_MARKER = ".turnkeybuild"

# Locate the models directory
MODELS_DIR = importlib.util.find_spec("turnkeyml_models").submodule_search_locations[0]


def rmdir(folder, excludes: Optional[List[str]] = None):
    """
    Remove the contents of a directory from the filesystem.
    If `<name>` is in `excludes`, the directory itself and the file named <name>
    are kept. Otherwise, the entire directory is removed.
    """

    # Use an empty list by default
    if excludes:
        excludes_to_use = excludes
    else:
        excludes_to_use = []

    if os.path.isdir(folder):
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            if file_path not in excludes_to_use:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)

        if excludes is None:
            shutil.rmtree(folder)

        return True

    else:
        return False


def get_all(path, exclude_path=False, file_type="state.yaml", recursive=True):
    if recursive:
        files = [
            os.path.join(dp, f)
            for dp, dn, filenames in os.walk(path)
            for f in filenames
            if file_type in f
        ]
    else:
        files = []
        dp, _, filenames = os.walk(path)
        for f in filenames:
            if file_type in f:
                files.append(os.path.join(dp, f))

    if exclude_path:
        files = [os.path.basename(f) for f in files]

    return files


def clean_file_name(script_path: str) -> str:
    # Trim the ".py" / ".onnx"
    return pathlib.Path(script_path).stem


class CacheError(exp.Error):
    """
    Raise this exception when the cache is being accessed incorrectly
    """


def _load_yaml(file) -> Dict:
    if os.path.isfile(file):
        with open(file, "r", encoding="utf8") as stream:
            return yaml.load(stream, Loader=yaml.FullLoader)
    else:
        return {}


def _save_yaml(dict: Dict, file):
    with open(file, "w", encoding="utf8") as outfile:
        yaml.dump(dict, outfile)


def print_yaml_file(file_path, description):
    if os.path.exists(file_path):
        with open(file_path, "r", encoding="utf-8") as file:
            printing.log_info(f"The {description} for {file_path} are:")
            print(file.read())
    else:
        raise CacheError(
            f"No {description} found at {file_path}. "
            "Try running `turnkey cache list` to see the builds in your build cache."
        )


def make_cache_dir(cache_dir: str):
    """
    Create the build and cache directories, and put hidden files in them
    to mark them as such.
    """

    os.makedirs(cache_dir, exist_ok=True)

    # File that indicates that the directory is a cache directory
    cache_file_path = os.path.join(cache_dir, CACHE_MARKER)
    open(cache_file_path, mode="w", encoding="utf").close()


def make_build_dir(cache_dir: str, build_name: str):
    """
    Create the build and cache directories, and put hidden files in them
    to mark them as such.
    """
    make_cache_dir(cache_dir)

    build_dir = build.output_dir(cache_dir, build_name)
    os.makedirs(build_dir, exist_ok=True)

    # File that indicates that the directory is a build directory
    build_file_path = os.path.join(build_dir, BUILD_MARKER)
    open(build_file_path, mode="w", encoding="utf").close()


def check_cache_dir(cache_dir: str):
    cache_file_path = os.path.join(cache_dir, CACHE_MARKER)
    if not os.path.isfile(cache_file_path):
        raise CacheError(
            f"{cache_dir} is not a cache directory generated by TurnkeyML. "
            "You can only clean, delete and generate reports for directories that "
            "have been generated by TurnkeyML. Set a different --cache-dir before "
            "trying again."
        )


def is_build_dir(cache_dir: str, build_name: str):
    build_dir = build.output_dir(cache_dir, build_name)
    build_file_path = os.path.join(build_dir, BUILD_MARKER)
    return os.path.isfile(build_file_path)


def clean_output_dir(cache_dir: str, build_name: str) -> None:
    """
    Delete all elements of the output directory that are not human readable
    """
    output_dir = os.path.join(cache_dir, build_name)
    if os.path.isdir(output_dir) and is_build_dir(cache_dir, build_name):
        output_dir = os.path.expanduser(output_dir)
    else:
        raise CacheError(f"No build found at {output_dir}")

    # Remove files that do not have an allowed extension
    allowed_extensions = (".txt", ".out", ".yaml", ".json")
    all_paths = glob.glob(f"{output_dir}/**/*", recursive=True)
    for path in all_paths:
        if os.path.isfile(path) and not path.endswith(allowed_extensions):
            os.remove(path)

    # Remove all empty folders
    for path in all_paths:
        if os.path.isdir(path):
            if len(os.listdir(path)) == 0:
                shutil.rmtree(path)


def get_available_scripts(search_dir: str):
    scripts = [
        f
        for f in os.listdir(search_dir)
        if os.path.isfile(os.path.join(search_dir, f)) and ".py" in f
    ]

    return scripts


def get_available_builds(cache_dir):
    """
    Get all of the build directories within the build cache
    located at `cache_dir`
    """

    check_cache_dir(cache_dir)

    builds = [
        pathlib.PurePath(build).name
        for build in os.listdir(os.path.abspath(cache_dir))
        if os.path.isdir(os.path.join(cache_dir, build))
        and is_build_dir(cache_dir, build)
    ]
    builds.sort()

    return builds


def print_available_builds(args):
    printing.log_info(f"Builds available in cache {args.cache_dir}:")
    builds = get_available_builds(args.cache_dir)
    printing.list_table(builds, num_cols=1)
    print()


def delete_builds(args):
    check_cache_dir(args.cache_dir)

    if args.delete_all:
        builds = get_available_builds(args.cache_dir)
    else:
        builds = [args.build_name]

    for build in builds:
        build_path = os.path.join(args.cache_dir, build)
        if is_build_dir(args.cache_dir, build):
            rmdir(build_path)
            printing.log_info(f"Deleted build: {build}")
        else:
            raise CacheError(
                f"No build found with name: {build}. "
                "Try running `turnkey cache list` to see the builds in your build cache."
            )


def clean_builds(args):
    check_cache_dir(args.cache_dir)

    if args.clean_all:
        builds = get_available_builds(args.cache_dir)
    else:
        builds = [args.build_name]

    for build in builds:
        if is_build_dir(args.cache_dir, build):
            clean_output_dir(args.cache_dir, build)
            printing.log_info(f"Removed the build artifacts from: {build}")
        else:
            raise CacheError(
                f"No build found with name: {build}. "
                "Try running `turnkey cache list` to see the builds in your build cache."
            )


def clean_build_name(build_name: str) -> str:
    """
    Remove hash from build name
    Build names have the format: <script_name>_<author>_hash
    """

    # Get everything except the trailing _<hash>
    return "_".join(build_name.split("_")[:-1])


def get_build_name(
    script_name: str, script_labels: Dict[str, List], model_hash: str = None
):
    """
    Create build name from script_name, labels and model hash
    """
    build_name = script_name
    if "author" in script_labels:
        build_name += f"_{script_labels['author'][0]}"
    if model_hash:
        build_name += f"_{model_hash}"
    return build_name


def get_builds_from_file(cache_dir, script):
    script_name = clean_file_name(script)
    script_labels = labels.load_from_file(script)
    all_builds_in_cache = get_available_builds(cache_dir)

    script_builds = [
        x
        for x in all_builds_in_cache
        if get_build_name(script_name, script_labels) == clean_build_name(x)
    ]

    return script_builds, script_name


class Keys:
    # Unique hash value that identifies the model+inputs+args
    # for a workload
    HASH = "hash"
    # Number of parameters in the model
    PARAMETERS = "parameters"
    # Histogram of ONNX operators used in the model
    ONNX_OPS_COUNTER = "onnx_ops_counter"
    # ONNX model info: IR version, opset, and size on disk (KiB)
    ONNX_MODEL_INFO = "onnx_model_information"
    # ONNX model input tensor dimensions
    ONNX_INPUT_DIMENSIONS = "onnx_input_dimensions"
    # List of all build stages in the Sequence
    SELECTED_SEQUENCE_OF_STAGES = "selected_sequence_of_stages"
    # Location of the most up-to-date ONNX file for this build. If the
    # build completed successfully, this is the final ONNX file.
    ONNX_FILE = "onnx_file"
    # MeasuredPerformance data for a benchmarked workload
    PERFORMANCE = "performance"
    # Runtime used for the benchmark
    RUNTIME = "runtime"
    # Device used for the benchmark
    DEVICE_TYPE = "device_type"
    # Name of the model
    MODEL_NAME = "model_name"
    # References the per-evaluation stats section
    EVALUATIONS = "evaluations"
    # Catch-all for storing a file's labels
    LABELS = "labels"
    # Author of the model
    AUTHOR = "author"
    # Class type of the model
    CLASS = "class"
    # Task the model is meant to perform
    TASK = "task"
    # Number of iterations used in benchmarking
    ITERATIONS = "iterations"
    # System information to keep track of DUT
    SYSTEM_INFO = "system_info"
    # Path to the built-in model script used as input
    MODEL_SCRIPT = "builtin_model_script"
    # Indicates status of the most recent build tool run: FunctionStatus
    BUILD_STATUS = "build_status"
    # Indicates status of the most recent benchmark tool run: FunctionStatus
    BENCHMARK_STATUS = "benchmark_status"
    # Indicates the match between the TorchScript IR graph and
    # the exported onnx model (verified with torch.onnx.verification)
    TORCH_ONNX_EXPORT_VALIDITY = "torch_export_validity"
    # Prefix for reporting the execution duration of a stage
    # In the report this will look like stage_duration:STAGE_NAME
    STAGE_DURATION = "stage_duration"
    # Prefix for reporting the execution status of a stage
    # In the report this will look like stage_status:STAGE_NAME
    STAGE_STATUS = "stage_status"


class Stats:
    def __init__(self, cache_dir: str, build_name: str, evaluation_id: str = None):
        output_dir = build.output_dir(cache_dir, build_name)

        self.file = os.path.join(output_dir, "turnkey_stats.yaml")
        self.evaluation_id = evaluation_id

        os.makedirs(output_dir, exist_ok=True)
        if not os.path.exists(self.file):
            initial = {Keys.EVALUATIONS: {}}
            _save_yaml(initial, self.file)

    @property
    def stats(self):
        return _load_yaml(self.file)

    def _set_key(self, dict, keys: List["str"], value):
        """
        Recursive approach to safely setting a key within any level of hierarchy
        in a dictionary. If a parent key of the desired key does not exist, create
        it and set it with an empty dictionary before proceeding.

        The end result is: dict[keys[0]][keys[1]]...[keys[-1]] = value
        """
        if len(keys) == 1:
            dict[keys[0]] = value

        else:
            if keys[0] not in dict.keys():
                dict[keys[0]] = {}

            self._set_key(dict[keys[0]], keys[1:], value)

    def save_model_stat(self, key: str, value):
        """
        Save statistics to an yaml file in the build directory
        """

        stats_dict = self.stats

        self._set_key(stats_dict, [key], value)

        _save_yaml(stats_dict, self.file)

    def save_model_eval_stat(self, key: str, value):
        stats_dict = self.stats

        self._set_key(stats_dict, [Keys.EVALUATIONS, self.evaluation_id, key], value)

        _save_yaml(stats_dict, self.file)

    def save_model_eval_sub_stat(self, parent_key: str, key: str, value):
        stats_dict = self.stats

        self._set_key(
            stats_dict, [Keys.EVALUATIONS, self.evaluation_id, parent_key, key], value
        )

        _save_yaml(stats_dict, self.file)

    @property
    def evaluation_stats(self):
        return self.stats[Keys.EVALUATIONS][self.evaluation_id]


def print_cache_dir(_=None):
    printing.log_info(f"The default cache directory is: {DEFAULT_CACHE_DIR}")


def print_models_dir(args=None):
    if args.verbose:
        printing.log_info(f"The models directory is: {MODELS_DIR}")
    else:
        print(MODELS_DIR)


def expand_inputs(input_paths: List[str]) -> List[str]:
    """
    Convert regular expressions in input paths
    into full file/dir paths (e.g., [*.py] -> [a.py, b.py] )

    This makes up for Windows not resolving wildcards on the command line
    """
    input_paths_expanded = sum(
        [glob.glob(f) for f in input_paths if "::" not in f], []
    ) + [f for f in input_paths if "::" in f]

    if not input_paths_expanded:
        raise exp.ArgError("No files that match your inputs could be found.")

    return input_paths_expanded


def rebase_cache_dir(input_path: str, build_name: str, new_cache_dir: str):
    """
    Rebase a turnkey build path onto a new turnkey cache directory.

    For example:
        1. You built a model in cache_A
        2. The turnkey_stats.yaml references cache_A/build_dir/onnx/model.onnx
        3. You move the build to cache_B
        4. You now call this function to get a new path, cache_B/build_dir/onnx/model.onnx

    This works by:
        1. Split on the first occurance of the build name: [cache_A, onnx/model.onnx]
        2. Prepend the new cache dir, as well as the build dir (since it was removed in step 1)
    """

    relative_input_path = input_path.split(build_name, 1)[1][1:]
    return os.path.join(new_cache_dir, build_name, relative_input_path)



"""
Functions that help us avoid importing tensorflow (TF), since that import
takes a very long time.

The test `if "tensorflow" in sys.modules:` checks to see if TF
has already been imported. This will always be true if someone is passing
a TF model since... that TF model had to come from somewhere :)

If TF hasn't already been imported, then there is no change that an object
is a TF instance, or TF is in any particular mode, or anything else, so
we can just return False on those checks.
"""

import sys
import inspect
from typing import List


def is_keras_model(model) -> bool:
    if "tensorflow" in sys.modules:
        return isinstance(model, sys.modules["tensorflow"].keras.Model)
    else:
        return False


def is_keras_tensor(tensor) -> bool:
    if "tensorflow" in sys.modules:
        return sys.modules["tensorflow"].is_tensor(tensor)
    else:
        return False


def is_executing_eagerly() -> bool:
    if "tensorflow" in sys.modules:
        return sys.modules["tensorflow"].executing_eagerly()
    else:
        return False


def type_is_tf_tensor(object) -> bool:
    if "tensorflow" in sys.modules:
        return object is sys.modules["tensorflow"].Tensor
    else:
        return False

def is_keras_subclass(obj_type) -> bool:
    if "tensorflow" in sys.modules:
        return issubclass(obj_type, sys.modules["tensorflow"].keras.Model)
    else:
        return False

def get_classes(module) -> List[str]:
    """
    Returns all classes within a module
    """
    return [y for x, y in inspect.getmembers(module, inspect.isclass)]

def get_transformers_activations() -> List:
    """
    We need this helper because `import transformers.activations` brings in `tensorflow`
    We can apply this helper because there is 0 chance of encountering a
    transformers activation if user code has not already imported transformers
    """
    if "transformers" in sys.modules:
        return get_classes(sys.modules["transformers"].activations)
    else:
        return []



# labels: test_group::turnkey name::efficientnet_b1 author::torch_hub task::Computer_Vision license::bsd-3-clause
from turnkeyml.parser import parse
import torch
from torchvision.models import EfficientNet_B1_Weights

torch.manual_seed(0)

# Parsing command-line arguments
pretrained, batch_size, num_channels, width, height = parse(
    ["pretrained", "batch_size", "num_channels", "width", "height"]
)


# Model and input configurations
model = torch.hub.load(
    "pytorch/vision:v0.13.1",
    "efficientnet_b1",
    weights=EfficientNet_B1_Weights.DEFAULT if pretrained else None,
)
model.eval()
inputs = {"x": torch.ones([batch_size, num_channels, width, height])}


# Call model
model(**inputs)
